# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""
This module allows you to make and combine a bunch of schedules, and setup the
optimizer so that it changes hyperparameter values based on the schedule. Highly
recommend you check out the `tutorials section <tutorials.html>`_ on this.
This is exposed automatically with::

   from k1lib.imports import *
   schedule.Fn # exposed
"""
import math, k1lib; import k1lib.cli as cli
import matplotlib.pyplot as plt, numpy as np
from itertools import accumulate
from k1lib.callbacks import Cbs, Callback
from typing import List, Callable, Union
__all__ = ["Fn", "linear", "smooth", "hump", "exp", "ParamScheduler"]
class Fn:
    def __init__(self, f:Callable[[float], float], param:str=None):
        """Creates a new schedule based on some custom function.
Example::

    s = schedule.Fn(lambda x: x**2)
    s(0.2) # returns 0.04

    # you can also use this as a decorator
    @schedule.Fn
    def s(x):
        return x**2

:param f: function (domain should always in [0, 1]), can be :class:`~k1lib.cli.modifier.op`
:param param: (optional) Parameter to schedule (e.g "lr") if using :class:`ParamScheduler`"""
        if isinstance(f, cli.op): f.op_solidify()
        self.f = f; self.param = param; self.progress = None
        self.domain = k1lib.Range(0, 1)
    def __call__(self, x:float):
        """Get the current value."""
        return self.f(x)
    def _startBatch(self, paramGroup:dict, progress:float):
        self.progress = progress
        paramGroup[self.param] = self(progress)
    @property
    def value(self): return self.f(self.progress)
    def __mul__(self, x): self.domain *= x; return self
    def __rmul__(self, x): self.domain *= x; return self
    def __truediv__(self, x): self.domain /= x; return self
    def __rtruediv__(self, x): self * (1.0/x); return self
    def __radd__(self, v):
        if isinstance(v, int): return self
        return NotImplemented
    def __add__(self, s:Union["Fn", str]) -> "Fn":
        """If given :class:`Fn`, then combines the 2 schedules together.
If it's a string, then sets the current param to it."""
        if isinstance(s, Fn): return CombinedSchedule(self, s)
        self.param = s; return self
    def iter(self, n:int):
        """Returns an n-step iterator evenly divided in range [0, 1].
Example::

    s = schedule.Fn(lambda x: x+2)
    list(s.iter(6)) # returns [2.0, 2.2, 2.4, 2.6, 2.8, 3.0]"""
        for e in np.linspace(0, 1, n): yield self(e)
    def modifyOutput(self, f:Callable[[float], float]) -> "Fn":
        """Returns a new :class:`Fn` that has its output modified.
Example::

    s = Fn(lambda x: x+2)
    s.modifyOutput(lambda x: x**2) # now s's function is (x+2)**2"""
        return Fn(lambda x: f(self.f(x)), self.param)
@k1lib.patch(Fn)
def __repr__(self):
    plt.figure(dpi=100); c = dict(color="tab:green")
    x = np.linspace(*self.domain, 1000); y = [self.f(x) for x in x]; plt.plot(x, y)
    y = self(0); plt.plot(0, y, "o", **c); plt.annotate("(0, {:.1e})".format(y), (0, y))
    y = self(1); plt.plot(1, y, "o", **c); plt.annotate("(1, {:.1e})".format(y), (1, y))
    x = self.progress
    if x is not None:
        blur = not (x in k1lib.Range(0.1, 0.9))
        y = self(x); plt.plot(x, y, "o", **c, alpha=(0.5 if blur else 1))
        if not blur: plt.annotate("({:.1e}, {:.1e})".format(x, y), (x, y))
    plt.show()
    return f"""'{self.param}' schedule. Can...
- s.progress: to get last recorded progress
- s.value: to get last recorded hyper parameter's value
- s(0.3): to get value of schedule at 30% progress"""
class CombinedSchedule(Fn):
    def __init__(self, s1, s2):
        split = s1.domain.stop / (s1.domain.delta + s2.domain.delta)
        s1r = k1lib.Range(0, split); s2r = k1lib.Range(split, 1)
        def f(x):
            if x < split: return s1.f(s1r.toUnit(x))
            else: return s2.f(s2r.toUnit(x))
        super().__init__(f, s1.param or s2.param)
def decorate(f:Callable[[float, float, float], float]) -> Fn:
    """Decorator, transforms f(low, high, x) to (low, high) -> f(x)."""
    def _f(low, high, param:str=None):
        return Fn(lambda x: f(low, high, x), param)
    return k1lib.wraps(f)(_f)
@decorate
def linear(low, high, x):
    """Sharply goes from low to high"""
    return low + x * (high - low)
@decorate
def smooth(low, high, x):
    """Smoothly goes from low to high"""
    return low + (high - low) * (1 + math.cos(math.pi * (1-x))) / 2
def hump(low, high, param:str=None):
    """Smoothly rises up (30%), then down (70%)"""
    return 0.3*smooth(0.8 * low + 0.2 * high, high) + 0.7*smooth(high, low, param)
_en4 = math.e**-3
@decorate
def exp(low, high, x):
    """Rises/drops quickly, then rate of change gets smaller and smaller"""
    return (math.exp(-x*4+1) - _en4) / (math.e - _en4) * (low - high) + high
@k1lib.patch(Cbs)
class ParamScheduler(Callback):
    """Schedules a param in parts of the network.

:param css: the selected parts of the network to schedule
:param schedules: (obvious)"""
    def __init__(self, css:str, *schedules:List[Fn]):
        super().__init__(); self.css = css
        for i, s in enumerate(schedules):
            if s.param is None: raise RuntimeError(f"Schedule {i} does not have associated parameter! Set with `s.param = 'lr'`.")
        self.schedules = {s.param:s for s in schedules}
        self.groupId = None; self.dependsOn = set("ProgressBar")
        self.initialized = False; self.prop = None
    def endRun(self):
        ":meta private:"
        self.initialized = False
    def __getstate__(self):
        answer = dict(self.__dict__)
        if "selector" in answer: del answer["selector"]
        return answer
    def startBatch(self):
        if self.l.model.training and self.groupId is not None:
            paramGroup = self.l.opt.param_groups[self.groupId]
            progress = self.l.progress
            for schedule in self.schedules.values():
                schedule._startBatch(paramGroup, progress)
    def __repr__(self):
        print(f"{self._reprHead}, css: \"{self.css}\", selector prop: \"{self.prop}\", schedules:")
        for schedule in self.schedules.values(): schedule.__repr__()
        return f"""Can...
- ps.schedules["lr"]: to get the schedule for a specific param
- ps.selector: to view the selected parameters
{self._reprCan}"""
@k1lib.patch(ParamScheduler, name="startRun")
def _startRun(self):
    if not self.initialized:
        # get all other ParamSchedulers
        pss = [cb for cb in self.l.cbs if isinstance(cb, ParamScheduler) and not cb.suspended]
        for i, ps in enumerate(pss):
            # make sure only 1 startRun is ran across all ParamSchedulers
            ps.initialized = True; ps.prop = f"_ps_{i}"
            ps.selector = k1lib.selector.select(self.l.model, ps.css)
            # sort pss based on depth, so that deeper ones gets accounted for first
            ps._depth = next(ps.selector.modules(ps.prop)).depth
        pss = sorted(pss, key=lambda ps: -ps._depth)
        # clear and add param groups
        self.l.opt.param_groups = []
        allParams = set(self.l.selector.nn.parameters())
        for ps in pss:
            params = set()
            for m in ps.selector.modules(ps.prop):
                for p in m.nn.parameters():
                    if p in allParams:
                        params.add(p); allParams.remove(p)
            if len(params) > 0:
                # so that we have a way to reference the group later on
                ps.groupId = len(self.l.opt.param_groups)
                self.l.opt.add_param_group({"prop": ps.prop, "css": ps.css, "params": list(params), **self.l.opt.defaults})
        self.l.opt.add_param_group({"prop": "rest", "css": "*", "params": list(allParams), **self.l.opt.defaults})
        for ps in pss:
            if ps.groupId is None: continue
            params = set(self.l.opt.param_groups[ps.groupId]["params"])
            def applyF(mS):
                mS.displayF = lambda s: "*" if any(p in params for p in s.directParams.values()) else ""
            ps.selector.apply(applyF)


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""
This is for functions that sort of changes the table
structure in a dramatic way. They're the core transformations
"""
from typing import List, Union, Iterator, Callable, Any, Tuple, Dict
from collections import defaultdict, Counter, deque
from k1lib.cli.init import patchDefaultDelim, BaseCli, oneToMany, T, Table, fastF, yieldT
import k1lib.cli as cli; from k1lib.cli.typehint import *
import itertools, numpy as np, k1lib; import matplotlib.pyplot as plt
try: import torch; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = True
__all__ = ["transpose", "reshape", "insert", "splitW", "splitC",
           "joinStreams", "joinStreamsRandom", "activeSamples",
           "table", "batched", "window", "groupBy",
           "insertColumn", "insertIdColumn",
           "expandE", "unsqueeze",
           "count", "permute", "accumulate", "AA_", "peek", "peekF",
           "repeat", "repeatF", "repeatFrom", "oneHot", "indexTable"]
settings = k1lib.settings.cli
class transpose(BaseCli):
    def __init__(self, dim1:int=0, dim2:int=1, fill=None):
        """Join multiple columns and loop through all rows. Aka transpose.
Example::

    # returns [[1, 4], [2, 5], [3, 6]]
    [[1, 2, 3], [4, 5, 6]] | transpose() | deref()
    # returns [[1, 4], [2, 5], [3, 6], [0, 7]]
    [[1, 2, 3], [4, 5, 6, 7]] | transpose(fill=0) | deref()

Multidimensional transpose works just like :meth:`torch.transpose` too::

    # returns (2, 7, 5, 3), but detected Tensor, so it will use builtin :meth:`torch.transpose`
    torch.randn(2, 3, 5, 7) | transpose(3, 1) | shape()
    # also returns (2, 7, 5, 3), but actually does every required computation. Can be slow if shape is huge
    torch.randn(2, 3, 5, 7) | deref(igT=False) | transpose(3, 1) | shape()

Can also work with numpy arrays::

    # returns (5, 3, 2)
    np.random.randn(2, 3, 5) | transpose(0, 2) | op().shape

Be careful with infinite streams, as transposing stream of shape (inf, 5) will
hang this operation! Either don't do it, or temporarily limit all infinite streams like
this::

    with settings.cli.context(inf=21):
        # returns (3, 21)
        [2, 1, 3] | repeat() | transpose() | shape()

Also be careful with empty streams, as you might not get any results at all::

    # returns [], as the last stream has no elements
    [[1, 2], [3, 4], []] | transpose() | deref()
    # returns [[1, 3, 0], [2, 4, 0]]
    [[1, 2], [3, 4], []] | transpose(fill=0) | deref()

:param fill: if not None, then will try to zip longest with this fill value"""
        super().__init__(); self.fill = fill
        self.d1 = min(dim1, dim2); self.d2 = max(dim1, dim2)
        self.normal = self.d1 == 0 and self.d2 == 1
    def _typehint(self, inp):
        if isinstance(inp, tArrayTypes):
            if inp.rank is None: return inp
            if inp.rank > max(self.d1, self.d2): return inp
            else: return tAny() # this case doesn't quite exist
        if self.d1 == 0 and self.d2 == 1:
            if isinstance(inp, tListIterSet):
                if isinstance(inp.child, tListIterSet):
                    return tIter(tList(inp.child.child))
        return tAny()
    def __ror__(self, it:Iterator[Iterator[T]]) -> Table[T]:
        d1 = self.d1; d2 = self.d2; fill = self.fill
        if isinstance(it, torch.Tensor): return it.transpose(d1, d2)
        if isinstance(it, np.ndarray):
            dims = list(range(len(it.shape)))
            temp = dims[d1]; dims[d1] = dims[d2]; dims[d2] = temp
            return it.transpose(dims)
        if d1 != 0 or d2 != 1: return it | cli.serial(*([transpose(fill=fill).all(i) for i in range(d1, d2)] + [transpose(fill=fill).all(i-1) for i in range(d2-1, d1, -1)]))
        if self.fill is None: return zip(*it)
        else: return itertools.zip_longest(*it, fillvalue=fill)
    @staticmethod
    def fill(fill="", dim1:int=0, dim2:int=1):
        """Convenience method to fill in missing elements of a table.
Example::

    # returns [[1, 2, 3], [4, 5, 0]]
    [[1, 2, 3], [4, 5]] | transpose.fill(0) | deref()
    # also returns [[1, 2, 3], [4, 5, 0]], demonstrating how it works underneath
    [[1, 2, 3], [4, 5]] | transpose(fill=0) | transpose(fill=0) | deref()"""
        return transpose(dim1, dim2, fill=fill) | transpose(dim1, dim2, fill=fill)
    @staticmethod
    def wrap(f, dim1:int=0, dim2:int=1, fill=None):
        """Wraps ``f`` around 2 :class:`transpose`, can be useful in combination with
:class:`k1lib.cli.init.mtmS`. Example::

    # returns [[1, 4, 3, 4], [8, 81, 10, 11]]
    [range(1, 5), range(8, 12)] | transpose.wrap(mtmS.f(apply(op()**2), 1)) | deref()
    # also returns [[1, 4, 3, 4], [8, 81, 10, 11]], demonstrating the typical way to do this
    [range(1, 5), range(8, 12)] | apply(op()**2, 1) | deref()

The example given is sort of to demonstrate this only. Most of the time, just use
:class:`~k1lib.cli.modifier.apply` with columns instead. But sometimes you need direct
access to a column, so this is how you can do it."""
        if not isinstance(f, BaseCli): f = cli.applyS(f)
        return transpose(dim1, dim2, fill) | f | transpose(dim1, dim2, fill)
#tOpt.clearPasses()
def oTranspose1(cs, ts, metadata): # `transpose() | transpose().all()` to `transpose() | apply(aS(torch.stack) | transpose())`
    tr, ap = cs; t = ts[0]
    if (not ap.normal) or (not tr.normal): return None
    if (not isinstance(ap.f, transpose)) or (not ap.f.normal): return None
    if isinstance(t, tListIterSet) and (isinstance(t.child, tTensor) or isinstance(t.child, tCollection) and isinstance(tLowest(*t.child.children), tTensor)):
        return [transpose(), cli.apply(cli.aS(torch.stack) | transpose())]
tOpt.addPass(oTranspose1, [transpose, cli.apply])
def _formStructure(it, dims, dimI):
    if dimI >= len(dims): return next(it)
    return [_formStructure(it, dims, dimI+1) for i in range(dims[dimI])]
class reshape(BaseCli):
    def __init__(self, *dims):
        """Reshapes the input stream into the desired shape.
Example::

    # returns [[0, 1, 2], [3, 4, 5]]
    range(6) | reshape(2, 3) | deref()
    # returns [[0, 1], [2, 3], [4, 5]]
    range(6) | reshape(3, 2) | deref()
    # returns [[0, 1], [2, 3], [4, 5]], stopped early
    range(100) | reshape(3, 2) | deref()
    # returns [[0, 1, 2], [3, 4, 5]], can leave out first dimension
    range(6) | reshape(-1, 3) | deref()
    # returns [[0, 1, 2]], won't include 2nd element, as it ran out of elements
    range(5) | reshape(-1, 3) | deref()
    # throws error, as it ran out of elements and can't fulfill the request
    range(6) | reshape(3, 3) | deref()

Unlike :meth:`torch.reshape`, the input piped into this has to be a simple iterator.
If you have a complex data structure with multiple dimensions, turn that into a simple
iterator with :class:`joinStreams` first, like this::

    # returns [[[0, 1, 2]], [[3, 4, 5]]]
    [[[0], [1]], [[2], [3]], [[4], [5]]] | joinStreams(2) | reshape(2, 1, 3) | deref()"""
        self.dims = dims
    def __ror__(self, it):
        it = iter(it)
        if self.dims[0] == -1:
            try:
                while True: yield _formStructure(it, self.dims, 1)
            except StopIteration: pass
        else:
            for i in range(self.dims[0]): yield _formStructure(it, self.dims, 1)
class insert(BaseCli):
    def __init__(self, elements, begin=True):
        """Join element into list.
Example::

    # returns [5, 2, 6, 8]
    [2, 6, 8] | insert(5) | deref()
    # returns [2, 6, 8, 5]
    [2, 6, 8] | insert(5, begin=False) | deref()

    # returns [[3, 1], 2, 6, 8]
    [2, 6, 8] | insert([3, 1]) | deref()

:param element: the element to insert"""
        super().__init__(); self.elements = elements; self.begin = begin; self.expand = False
    def __ror__(self, it:Tuple[T, Iterator[T]]) -> Iterator[T]:
        elements = self.elements; it = iter(it)
        if self.begin: yield elements; yield from it
        else: yield from it; yield elements
class splitW(BaseCli):
    def __init__(self, *weights:List[float]):
        """Splits elements into multiple weighted lists. If no weights are provided,
then automatically defaults to [0.8, 0.2]. Example::

    # returns [[0, 1, 2, 3, 4, 5, 6, 7], [8, 9]]
    range(10) | splitW(0.8, 0.2) | deref()
    # same as the above
    range(10) | splitW() | deref()

This also works with array types::

    torch.randn(100, 3) | splitW() # returns 2 tensors with shapes (80, 3) and (20, 3)

See also: :class:`splitC`"""
        super().__init__();
        if len(weights) == 0: weights = [0.8, 0.2]
        self.weights = np.array(weights)
    def __ror__(self, it):
        try: it[0]; len(it)
        except: it = list(it)
        ws = self.weights; c = 0; ws = (ws * len(it) / ws.sum()).astype(int)
        for w in ws[:-1]: yield it[c:c+w]; c += w
        yield it[c:]
class splitC(BaseCli):
    def __init__(self, *checkpoints:List[float]):
        """Splits elements into multiple checkpoint-delimited lists.
Example::

    # returns [[0, 1], [2, 3, 4], [5, 6, 7, 8, 9]]
    range(10) | splitC(2, 5) | deref()
    # returns ['01', '234', '56789']
    "0123456789" | splitC(2, 5) | deref()

Here, you're specifying 2 checkpoints, 2 and 5, so it will split
the list up into 3 sections. First section is 0-2, second section
is 2-5, third section is 5-end. You can pass in fractional
checkpoints too::

    # returns [[0, 1], [2, 3, 4, 5], [6, 7, 8, 9]]
    range(10) | splitC(0.2, 0.6) | deref()

This cli might be unintuitive to remember, so if you want to just
split it up into 2 parts, check out :meth:`~k1lib.cli.filt.head.split`.

If you want to split things up by weighted length, check
out :class:`splitW`"""
        self.checkpoints = checkpoints | cli.aS(np.array)
        self.intMode = checkpoints | cli.apply(lambda x: int(x) == x) | cli.aS(all)
    def __ror__(self, it):
        try: it[0]; len(it)
        except: it = list(it)
        cs = self.checkpoints
        if not self.intMode: cs = (cs * len(it)).astype(int)
        cs = sorted(cs); yield it[:cs[0]]
        for i in range(len(cs)-1): yield it[cs[i]:cs[i+1]]
        yield it[cs[-1]:]
class joinStreams(BaseCli):
    def __init__(self, dims=1):
        """Joins multiple streams.
Example::

    # returns [1, 2, 3, 4, 5]
    [[1, 2, 3], [4, 5]] | joinStreams() | deref()
    # returns [[0, 1], [2], [3, 4, 5], [6, 7, 8], [], [9, 10]]
    [[[0, 1], [2], [3, 4, 5]], [[6, 7, 8], [], [9, 10]]] | joinStreams() | deref()
    # returns [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    [[[0, 1], [2], [3, 4, 5]], [[6, 7, 8], [], [9, 10]]] | joinStreams(2) | deref()

If you pass in :class:`numpy.ndarray` or :class:`torch.Tensor`, then it will
automatically use the C-accelerated version, like this::

    # returns Tensor with shape (6, 4)
    torch.randn(2, 3, 4) | joinStreams()
    # returns array with shape (6, 4)
    np.random.randn(2, 3, 4) | joinStreams()

Sometimes, you may want to impose some dimensional structure after joining all streams
together, which :class:`reshape` does."""
        if dims < 1: raise AttributeError(f"`dims` ({dims}) can't be less than 1, as it doesn't make any sense!")
        self.multi = cli.serial(*(joinStreams() for d in range(dims))) if dims > 1 else None
    def __ror__(self, streams:Iterator[Iterator[T]]) -> Iterator[T]:
        if self.multi != None:
            return streams | self.multi
            def gen(): yield from streams | self.multi
            return gen()
        else:
            if isinstance(streams, settings.arrayTypes): return streams.reshape(-1, *streams.shape[2:])
            else:
                def gen():
                    for stream in streams: yield from stream
                return gen()
def probScale(ps, t): # t from 0 -> 1, for typical usage
    l = np.log(ps); avg = l.mean()
    a = (l-avg)*t+avg; a -= a.max()
    ans = np.exp(a); return ans/ans.sum()
import random
def rand(n, ps=None):
    if ps is None:
        while True: yield random.randrange(n)
    else:
        while True: yield from np.random.choice(n, size=100, p=ps)
class joinStreamsRandom(BaseCli):
    def __init__(self, alpha=0, ps=None):
        """Join multiple streams randomly. If any streams runs out, then quits. If
any stream yields :data:`~k1lib.cli.init.yieldT`, then just ignores that result and
continue. Could be useful in active learning. Example::

    # could return [0, 1, 10, 2, 11, 12, 13, ...], with max length 20, typical length 18
    [range(0, 10), range(10, 20)] | joinStreamsRandom() | deref()
    
    stream2 = [[-5, yieldT, -4, -3], yieldT | repeat()] | joinStreams()
    # could return [-5, -4, 0, -3, 1, 2, 3, 4, 5, 6], demonstrating yieldT
    [range(7), stream2] | joinStreamsRandom() | deref()

By default, all streams are treated equally, and are yielded with equal probabilities.
However, you can tweak these probabilities a little bit, to your liking. This is
controlled by the parameter ``alpha``:

.. image:: ../images/probScale.png

If ``alpha`` is 0, then all probabilities will be the same. If ``alpha`` is 1,
then all probabilities are proportional to the length of the input stream. The
original intention was to vary ``alpha`` just from 0 to 1, but it can actually
be of any number::

    [range(0, 10), range(10, 100)] | joinStreamsRandom(0) | shape(0) # returns around 21, because it favors both streams equally
    [range(0, 10), range(10, 100)] | joinStreamsRandom(1) | shape(0) # returns around 90, because it favors the second array 9x more
    [range(0, 10), range(10, 100)] | joinStreamsRandom(100) | shape(0) # returns 90, because it highly favors the second array
    [range(0, 10), range(10, 100)] | joinStreamsRandom(-100) | shape(0) # returns 10, because it highly favors the first array

:param alpha: if not zero, does a weighted joining, instead of totally uniform probability
:param ps: if specified, use these probabilities, else try to determine from the lengths of the input streams"""
        super().__init__(); self.alpha = alpha; self.ps = ps
    def __ror__(self, streams:Iterator[Iterator[T]]) -> Iterator[T]:
        alpha = self.alpha; ps = self.ps; streams = list(streams); nStreams = len(streams)
        if alpha != 0:
            if ps is None:
                try: ps = np.array([len(st) for st in streams])
                except:
                    streams = [list(st) for st in streams]
                    ps = np.array([len(st) for st in streams])
            else: ps = np.array(list(ps))*1.0
        else: ps = np.array([1/nStreams]*nStreams)
        ps = probScale(ps/ps.sum(), alpha)
        streams = [iter(st) for st in streams]
        try:
            for streamIdx in rand(len(streams), ps):
                o = next(streams[streamIdx])
                if not o is yieldT: yield o # "not is" to fix numpy `==`
        except StopIteration: pass
class activeSamples(BaseCli):
    def __init__(self, limit:int=100, p:float=0.95):
        """Yields active learning samples.
Example::

    o = activeSamples()
    ds = range(10) # normal dataset
    ds = [o, ds] | joinStreamsRandom() # dataset with active learning capability
    next(ds) # returns 0
    next(ds) # returns 1
    next(ds) # returns 2
    o.append(20)
    next(ds) # can return     3     or 20
    next(ds) # can return (4 or 20) or 4

So the point of this is to be a generator of samples. You can define your dataset
as a mix of active learning samples and standard samples. Whenever there's a data
point that you want to focus on, you can add it to ``o`` and it will eventially yield
it.

.. warning::

    It might not be a good idea to set param ``limit`` to higher numbers than
    100. This is because, the network might still not understand a wrong sample
    after being shown multiple times, and will keep adding that wrong sample
    back in, distracting it from other samples, and reduce network's accuracy
    after removing active learning from it.
    
    If ``limit`` is low enough (from my testing, 30-100 should be fine), then
    old wrong samples will be kicked out, allowing for a fresh stream of wrong
    samples coming in, and preventing the problem above. If you found that
    removing active learning makes the accuracy drops dramatically, then try
    decreasing the limit.

:param limit: max number of active samples. Discards samples if number of samples
    is over this.
:param p: probability of actually adding the samples in"""
        super().__init__(); self.p = p
        self.samples = deque([], limit)
    def append(self, item):
        """Adds 1 sample."""
        if random.random() < self.p: self.samples.append(item)
    def extend(self, items):
        """Adds multiple samples."""
        for item in items: self.append(item)
    def __iter__(self):
        samples = self.samples
        while True:
            if len(samples) == 0: yield yieldT
            else: yield samples.popleft()
def table(delim:str=None):
    """Basically ``op().split(delim).all()``. This exists because this is used
quite a lot in bioinformatics. Example::

    # returns [['a', 'bd'], ['1', '2', '3']]
    ["a|bd", "1|2|3"] | table("|") | deref()"""
    return cli.op().split(patchDefaultDelim(delim)).all()
def _batch(it, bs, includeLast):
    l = []; it = iter(it)
    try:
        while True:
            for i in range(bs): l.append(next(it))
            yield l; l = []
    except StopIteration:
        if includeLast and len(l) > 0: yield l
def _batchRange(it, bs, includeLast):
    start, stop, step = it.start, it.stop, it.step
    lastCur = start; cur = lastCur + bs*step
    while cur <= stop:
        yield range(lastCur, cur, step)
        lastCur = cur; cur += bs*step
    if includeLast and lastCur < stop:
        yield range(lastCur, stop, step)
def _batchSliceable(it, bs, includeLast):
    cur = 0; n = len(it)
    while (cur+1)*bs <= n:
        yield it[cur*bs:(cur+1)*bs]
        cur += 1
    if includeLast:
        ans = it[cur*bs:(cur+1)*bs]
        if len(ans) > 0: yield ans
class batched(BaseCli):
    def __init__(self, bs=32, includeLast=False):
        """Batches the input stream.
Example::

    # returns [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    range(11) | batched(3) | deref()
    # returns [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]
    range(11) | batched(3, True) | deref()
    # returns [[0, 1, 2, 3, 4]]
    range(5) | batched(float("inf"), True) | deref()
    # returns []
    range(5) | batched(float("inf"), False) | deref()
    
Can work well and fast with :class:`torch.Tensor` and :class:`numpy.ndarray`::

    # both returns torch.Tensor of shape (2, 3, 4, 5)
    torch.randn(6, 4, 5) | batched(3)
    torch.randn(7, 4, 5) | batched(3)

Also, if input is a :class:`range`, then to save time, a bunch of other
ranges will be returned, instead of a bunch of lists, for performance::

    # returns [range(0, 3), range(3, 6), range(6, 9)]
    range(11) | batched(3) | toList()
"""
        super().__init__(); self.bs = bs; self.includeLast = includeLast
    def __ror__(self, it):
        bs = self.bs; includeLast = self.includeLast
        if bs == float("inf"):
            if includeLast: return [it]
            return []
        if not includeLast and isinstance(it, k1lib.settings.cli.arrayTypes):
            n = it.shape[0] // bs; it = it[:n*bs]
            return it.reshape(n, bs, *it.shape[1:])
        if isinstance(it, range): return _batchRange(it, bs, includeLast)
        try: it[0]; len(it); return _batchSliceable(it, bs, includeLast)
        except: return _batch(it, bs, includeLast)
nothing = object()
class window(BaseCli):
    def __init__(self, n, newList=False, pad=nothing):
        """Slides window of size n forward and yields the windows.
Example::

    # returns [[0, 1, 2], [1, 2, 3], [2, 3, 4]]
    range(5) | window(3) | deref()
    # returns [[0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, None], [4, None, None]]
    range(5) | window(3, pad=None) | deref()

If you are doing strange transformations to the result, like
transposing it, then it might complain that the internal deque
(double-ended queue) mutated during iteration. In that case,
then set ``newList`` to True. It's not True by default because
multiple lists will be created, all of which needs memory
allocation, which will be slower::

    # takes 15ms
    range(100000) | window(100) | ignore()
    # takes 48ms, because of allocating lists
    range(100000) | window(100) | ignore()

:param n: size of the window
:param newList: whether to create a new list out of every window or
    not. If False (default), less robust but faster. If True, more
    robust but slower
:param pad: whether to pad the output stream on the end, so that it
    has the same number of elements as the input stream or not"""
        self.n = n; self.listF = (lambda x: list(x)) if newList else (lambda x: iter(x)); self.pad = pad
    def __ror__(self, it):
        n = self.n; pad = self.pad; q = deque([], n); listF = self.listF
        for e in it:
            q.append(e)
            if len(q) == n: yield listF(q); q.popleft()
        if pad is not nothing:
            for i in range(n-1):
                q.append(pad)
                yield listF(q); q.popleft()
class groupBy(BaseCli):
    def __init__(self, column:int):
        """Groups table by some column.
Example::

    [[2.3, 5],
     [3.4, 2],
     [4.5, 2],
     [5.6, 5],
     [6.7, 1]] | groupBy(1) | deref()

This returns::

    [[[2.3, 5],
      [5.6, 5]],
     [[3.4, 2],
      [4.5, 2]],
     [[6.7, 1]]]

Should have O(n log(n)) time complexity

See also: :class:`~k1lib.cli.grep.grep`

:param column: which column to group by"""
        self.column = column
    def __ror__(self, it):
        it = it | cli.deref(2); c = self.column
        it = it | cli.sort(c, False)
        it = iter(it); a = [next(it)]; v = a[0][c]
        try:
            while True:
                e = next(it)
                if e[c] == v: a.append(e)
                else: yield a; a = [e]; v = a[0][c]
        except StopIteration:
            if len(a) > 0: yield a
class insertColumn(BaseCli):
    def __init__(self, column:List[T], begin=True, fill=""):
        """Inserts a column at beginning or end.
Example::

    # returns [['a', 1, 2], ['b', 3, 4]]
    [[1, 2], [3, 4]] | insertColumn(["a", "b"]) | deref()
    # returns [[1, 2, 'a'], [3, 4, 'b']]
    [[1, 2], [3, 4]] | insertColumn(["a", "b"], begin=False) | deref()"""
        self.column = column; self.begin = begin; self.fill = fill
    def __ror__(self, it):
        return it | transpose(fill=self.fill) | insert(self.column, begin=self.begin) | transpose(fill=self.fill)
def insertIdColumn(table=False, begin=True):
    """Inserts an id column at the beginning (or end).
Example::

    # returns [[0, 'a', 2], [1, 'b', 4]]
    [["a", 2], ["b", 4]] | insertIdColumn(True) | deref()
    # returns [[0, 'a'], [1, 'b']]
    "ab" | insertIdColumn()

:param table: if False, then insert column to an Iterator[str], else treat
    input as a full fledged table"""
    def a(it):
        if table:
            if begin: return ([i, *e] for i, e in enumerate(it))
            else: return ([*e, i] for i, e in enumerate(it))
        else:
            if begin: return ([i, e] for i, e in enumerate(it))
            else: return ([e, i] for i, e in enumerate(it))
    return cli.aS(a)
class expandE(BaseCli):
    def __init__(self, f:Callable[[T], List[T]], column:int):
        """Expands table element to multiple columns.
Example::

    # returns [['abc', 3, -2], ['de', 2, -5]]
    [["abc", -2], ["de", -5]] | expandE(lambda e: (e, len(e)), 0) | deref()

:param f: Function that transforms 1 row element to multiple elements"""
        super().__init__(fs=[f]); self.f = f; self.column = column
    def __ror__(self, it):
        f = self.f; c = self.column
        def gen(row):
            for i, e in enumerate(row):
                if i == c: yield from f(e)
                else: yield e
        return (gen(row) for row in it)
def unsqueeze(dim:int=0):
    """Unsqueeze input iterator.
Example::

    t = [[1, 2], [3, 4], [5, 6]]
    # returns (3, 2)
    t | shape()
    # returns (1, 3, 2)
    t | unsqueeze(0) | shape()
    # returns (3, 1, 2)
    t | unsqueeze(1) | shape()
    # returns (3, 2, 1)
    t | unsqueeze(2) | shape()

Behind the scenes, it's really just ``wrapList().all(dim)``, but the "unsqueeze" name
is a lot more familiar. Also note that the inverse operation "squeeze" is sort of
``item().all(dim)``, if you're sure that this is desirable::

    t = [[1, 2], [3, 4], [5, 6]]
    # returns (3, 2)
    t | unsqueeze(1) | item().all(1) | shape()"""
    return cli.wrapList().all(dim)
def oUnsqueeze(cs, ts, metadata): # reminder: if change this then also change the example in llvm.rst
    a = cs[0]; t = ts[0]; i = 0;
    if not isinstance(t, tArrayTypes): return None
    while isinstance(a, cli.apply) and a.normal: i += 1; a = a.f
    if not isinstance(a, cli.wrapList): return None
    t = t.__class__(t.child, t.rank+1 if t.rank is not None else None)
    if isinstance(t, tNpArray): return [cli.aS(lambda x: np.expand_dims(x, i)).hint(t)]
    else: return [cli.aS(lambda x: x.unsqueeze(i)).hint(t)]
tOpt.addPass(oUnsqueeze, [cli.apply], 4)
class count(BaseCli):
    def __init__(self):
        """Finds unique elements and returns a table with [frequency, value, percent]
columns. Example::

    # returns [[1, 'a', '33%'], [2, 'b', '67%']]
    ['a', 'b', 'b'] | count() | deref()"""
        super().__init__()
    def _typehint(self, inp):
        i = tAny()
        if isinstance(inp, tListIterSet): i = inp.child
        return tIter(tCollection(int, i, str))
    def __ror__(self, it:Iterator[str]):
        it = it | cli.apply(lambda row: (tuple(row) if isinstance(row, list) else row))
        c = Counter(it); s = sum(c.values())
        for k, v in c.items(): yield [v, k, f"{round(100*v/s)}%"]
    @staticmethod
    def join():
        """Joins multiple counts together.
Example::

    # returns [[2, 'a', '33%'], [4, 'b', '67%']]
    ['a', 'b', 'b'] | repeat(2) | applyMp(count() | deref()) | count.join() | deref()

This is useful when you want to get the count of a really long list/iterator using multiple cores"""
        def inner(counts):
            values = defaultdict(lambda: 0)
            for _count in counts:
                if _count is None: continue
                for v, k, *_ in _count:
                    values[k] += v
            s = values.values() | cli.toSum()
            for k, v in values.items(): yield [v, k, f"{round(100*v/s)}%"]
        return cli.applyS(inner)
def _permuteGen(row, pers):
    row = list(row); return (row[i] for i in pers)
class permute(BaseCli):
    def __init__(self, *permutations:List[int]):
        """Permutes the columns. Acts kinda like :meth:`torch.Tensor.permute`.
Example::

    # returns [['b', 'a'], ['d', 'c']]
    ["ab", "cd"] | permute(1, 0) | deref()"""
        super().__init__(); self.permutations = permutations
    def __ror__(self, it:Iterator[str]):
        p = self.permutations
        for row in it: yield _permuteGen(row, p)
class accumulate(BaseCli):
    def __init__(self, columnIdx:int=0, avg=False):
        """Groups lines that have the same row[columnIdx], and
add together all other columns, assuming they're numbers. Example::

    # returns [['a', 10.5, 9.5, 14.5], ['b', 1.1, 2.2, 3.3]]
    [["a", 1.1, 2.2, 3.4],
     ["a", 1.1, 2.2, 7.8],
     ["a", 8.3, 5.1, 3.3],
     ["b", 1.1, 2.2, 3.3]] | accumulate(0) | deref()

:param columnIdx: common column index to accumulate
:param avg: calculate average values instead of sum"""
        super().__init__(); self.columnIdx = columnIdx; self.avg = avg
        self.dict = defaultdict(lambda: defaultdict(lambda: 0))
        self.keyAppearances = defaultdict(lambda: 0)
    def __ror__(self, it:Iterator[str]):
        for row in it:
            row = list(row); key = row[self.columnIdx]
            row.pop(self.columnIdx); self.keyAppearances[key] += 1
            for i, e in enumerate(row):
                try: self.dict[key][i] += float(e)
                except: self.dict[key][i] = e
        for key, cols in self.dict.items():
            ncol = len(cols)
            if self.avg:
                for i, col in enumerate(cols):
                    if isinstance(col, (int, float)): cols[i] /= self.keyAppearances[key]
            elems = list(cols.values()); elems.insert(self.columnIdx, key); yield elems
class AA_(BaseCli):
    def __init__(self, *idxs:List[int], wraps=False):
        """Returns 2 streams, one that has the selected element, and the other
the rest. Example::

    # returns [5, [1, 6, 3, 7]]
    [1, 5, 6, 3, 7] | AA_(1)
    # returns [[5, [1, 6, 3, 7]]]
    [1, 5, 6, 3, 7] | AA_(1, wraps=True)

You can also put multiple indexes through::

    # returns [[1, [5, 6]], [6, [1, 5]]]
    [1, 5, 6] | AA_(0, 2)

If you don't specify anything, then all indexes will be sliced::

    # returns [[1, [5, 6]], [5, [1, 6]], [6, [1, 5]]]
    [1, 5, 6] | AA_()

As for why the strange name, think of this operation as "AĀ". In statistics,
say you have a set "A", then "not A" is commonly written as A with an overline
"Ā". So "AA\_" represents "AĀ", and that it first returns the selection A.

:param wraps: if True, then the first example will return [[5, [1, 6, 3, 7]]]
    instead, so that A has the same signature as Ā"""
        super().__init__(); self.idxs = idxs; self.wraps = wraps
    def __ror__(self, it:List[Any]) -> List[List[List[Any]]]:
        super().__ror__(it); idxs = self.idxs; it = list(it)
        if len(idxs) == 0: idxs = range(len(it))
        def gen(idx):
            return [it[idx], [v for i, v in enumerate(it) if i != idx]]
        if not self.wraps and len(idxs) == 1: return gen(idxs[0])
        return [gen(idx) for idx in idxs]
class peek(BaseCli):
    def __init__(self):
        """Returns (firstRow, iterator). This sort of peaks at the first row,
to potentially gain some insights about the internal formats. The returned
iterator is not tampered. Example::

    e, it = iter([[1, 2, 3], [1, 2]]) | peek()
    print(e) # prints "[1, 2, 3]"
    s = 0
    for e in it: s += len(e)
    print(s) # prints "5", or length of 2 lists

You kinda have to be careful about handling the ``firstRow``, because you might
inadvertently alter the iterator::

    e, it = iter([iter(range(3)), range(4), range(2)]) | peek()
    e = list(e) # e is [0, 1, 2]
    list(next(it)) # supposed to be the same as `e`, but is [] instead

The example happens because you have already consumed all elements of the first
row, and thus there aren't any left when you try to call ``next(it)``."""
        super().__init__()
    def __ror__(self, it:Iterator[T]) -> Tuple[T, Iterator[T]]:
        it = iter(it); sentinel = object(); row = next(it, sentinel)
        if row == sentinel: return None, []
        def gen(): yield row; yield from it
        return row, gen()
class peekF(BaseCli):
    def __init__(self, f:Union[BaseCli, Callable[[T], T]]):
        r"""Similar to :class:`peek`, but will execute ``f(row)`` and
return the input Iterator, which is not tampered. Example::

    it = lambda: iter([[1, 2, 3], [1, 2]])
    # prints "[1, 2, 3]" and returns [[1, 2, 3], [1, 2]]
    it() | peekF(lambda x: print(x)) | deref()
    # prints "1\n2\n3"
    it() | peekF(headOut()) | deref()"""
        super().__init__(fs=[f]); self.f = f
    def __ror__(self, it:Iterator[T]) -> Iterator[T]:
        it = iter(it); sentinel = object(); row = next(it, sentinel)
        if row == sentinel: return []
        def gen(): yield row; yield from it
        self.f(row); return gen()
class repeat(BaseCli):
    def __init__(self, limit:int=None):
        """Yields a specified amount of the passed in object. If you intend
to pass in an iterator, then make a list out of it first, as second copy of
iterator probably won't work as you will have used it the first time. Example::

    # returns [[1, 2, 3], [1, 2, 3], [1, 2, 3]]
    [1, 2, 3] | repeat(3) | toList()

:param repeat: if None, then repeats indefinitely"""
        super().__init__(); self.limit = limit
    def _typehint(self, inp): return tIter(inp)
    def __ror__(self, o:T) -> Iterator[T]:
        limit = self.limit if self.limit != None else k1lib.settings.cli.inf
        for i in itertools.count():
            if i >= limit: break
            yield o
def repeatF(f, limit:int=None, **kwargs):
    """Yields a specified amount generated by a specified function.
Example::

    # returns [4, 4, 4]
    repeatF(lambda: 4, 3) | toList()
    # returns 10
    repeatF(lambda: 4) | head() | shape(0)

    f = lambda a: a+2
    # returns [8, 8, 8]
    repeatF(f, 3, a=6) | toList()

:param limit: if None, then repeats indefinitely
:param kwargs: extra keyword arguments that you can pass into the function

See also: :class:`repeatFrom`"""
    f = fastF(f); limit = limit if limit != None else k1lib.settings.cli.inf
    if len(kwargs) == 0:
        for i in itertools.count():
            if i >= limit: break
            yield f()
    else:
        for i in itertools.count():
            if i >= limit: break
            yield f(**kwargs)
class repeatFrom(BaseCli):
    def __init__(self, limit:int=None):
        """Yields from a list. If runs out of elements, then do it again for
``limit`` times. Example::

    # returns [1, 2, 3, 1, 2]
    [1, 2, 3] | repeatFrom() | head(5) | deref()
    # returns [1, 2, 3, 1, 2, 3]
    [1, 2, 3] | repeatFrom(2) | deref()

.. note::

    For advanced users who wants to modify the resulting stream mid-way, read this section

    Because this reuses elements inside the input iterator, it's necessary
    that the input feels like a list and not an iterator. So in order to make
    this work::

        # returns [1, 2, 3, 1, 2, 3]
        iter([1, 2, 3]) | repeatFrom(2) | deref()

    It's necessary to turn the input iterator into a list. However, sometimes you
    may want to update the input iterator values, so as to make things extra
    dynamic, like this::

        l = [1, 2, 3]
        def g(): yield from l; yield from l
        def h():
            for i, e in enumerate(g()):
                if i == 3: l.append(5) # modifies the list mid-way
                yield e

        h() | deref() # returns [1, 2, 3, 1, 2, 3, 5]

    But if you do this, it wouldn't work::

        l = [1, 2, 3]
        def h():
            for i, e in enumerate(iter(l) | repeatFrom(2)):
                if i == 3: l.append(5)
                yield e
        h() | deref() # returns [1, 2, 3, 1, 2, 3]

    This is because internally, :class:`repeatFrom` turns the iterator into a
    list, and continues yielding from that list, and thus won't use the updated
    values. To do it, you have to make the input feels like a list (can get length)::

        l = [1, 2, 3]
        def h():
            for i, e in enumerate(l | repeatFrom(2)):
                if i == 3: l.append(5)
                yield e
        h() | deref() # returns [1, 2, 3, 1, 2, 3, 5]

:param limit: if None, then repeats indefinitely"""
        super().__init__(); self.limit = limit
    def _typehint(self, inp):
        i = tAny()
        if isinstance(inp, tListIterSet): i = inp.child
        if isinstance(inp, tArrayTypes): i = inp
        return tIter(i)
    def __ror__(self, it:Iterator[T]) -> Iterator[T]:
        try: len(it)
        except: it = list(it)
        limit = self.limit or k1lib.settings.cli.inf
        for i in itertools.count():
            if i >= limit: break
            yield from it
def oneHotRow(i, n): ans = [0]*n; ans[i] = 1; return ans
class oneHot(BaseCli):
    _groups = {}
    def __init__(self, col, n:int=0, group:str=None, sep:bool=False):
        """One-hot encode some column in a table.
Example::

    a = [
     [1, 2, "A"],
     [3, 4, "B"],
     [5, 6, "C"]]
    b = [
     [7, 8, "A"],
     [9, 10, "B"],
     [11, 12, "B"]]
    [*a, *b] | oneHot(2) | deref()
    [*a, *b] | oneHot(2, 3, "abcd") | deref()

Last 2 statements both return this::

    [[1, 2, 1, 0, 0],
     [3, 4, 0, 1, 0],
     [5, 6, 0, 0, 1],
     [7, 8, 1, 0, 0],
     [9, 10, 0, 1, 0],
     [11, 12, 0, 1, 0]]

You can also separate the encoded column out like this::

    [*a, *b] | oneHot(2, sep=True) | deref()

Which returns this::

    [[1, 2, [1, 0, 0]],
     [3, 4, [0, 1, 0]],
     [5, 6, [0, 0, 1]],
     [7, 8, [1, 0, 0]],
     [9, 10, [0, 1, 0]],
     [11, 12, [0, 1, 0]]]

The natural way to do this is to use with without ``n`` and ``group`` parameters.
But sometimes, your one hot encoding is spreaded across multiple datasets in
multiple dataloaders, and so the order and length of the encoding might not be
the same, which will mess up your training process.

That's why, you can specify ``group``, which will share encoding information
across all :class:`oneHot` clis that have the same group name. If you choose to
do this then you have to also specify what's the size of the encoding, because
the cli can't really infer the size when it potentially has not seen all the data
right?

:param col: which column one hot encode and expand into
:param n: (optional) total number of different elements
:param group: (optional) group name
:param sep: (optional) whether to separate the variable out into its own list"""
        self.col = col; self.n = n; self.group = group; self.sep = sep
        if (n != 0 and group is not None) and (n == 0 or group is None):
            raise Exception("You have to specify both `n` and `group` at the same time if you want to use them")
        if group is not None:
            if group not in oneHot._groups:
                oneHot._groups[group] = dict()
            self.d = oneHot._groups[group]
        else: self.d = dict()
    def _typehint(self, inp):
         # TODO
        if isinstance(inp, tListIterSet):
            if isinstance(inp.child, tListIterSet):
                pass
            elif isinstance(inp.child, tListIterSet):
                pass
            pass
        return tIter(tAny())
    def __ror__(self, it):
        c = self.col; d = self.d; n = self.n; sep = self.sep
        if n == 0:
            it = it | cli.deref(2); n = it | cli.cut(c) | cli.aS(set) | cli.shape(0)
        for row in it:
            e = row[c]
            try: e[0]; len(e)
            except: e = list(e)
            if e not in d: d[e] = oneHotRow(len(d), n)
            if sep: yield [*row[:c], d[e], *row[c+1:]]
            else:   yield [*row[:c], *d[e], *row[c+1:]]
        return
        _d = it | cli.cut(c) | cli.aS(set) | cli.sort(None, False) | cli.deref(); n = len(_d)
        _d = _d | insertIdColumn(begin=False) | cli.apply(cli.aS(oneHotRow, n), 1) | transpose() | cli.toDict()
        for row in it: yield [*row[:c], *_d[row[c]], *row[c+1:]]
class indexTable(BaseCli):
    def __init__(self, *cols):
        """Indexes a table by some columns.
Example::

    a = [
        [0, 3, 0.1],
        [0, 4, 0.2],
        [1, 3, 0.3],
        [1, 4, 0.4],
    ]
    # returns {3: [[0, 3, 0.1], [1, 3, 0.3]], 4: [[0, 4, 0.2], [1, 4, 0.4]]}
    a | indexTable(1)
    # returns {0: [[0, 3, 0.1], [0, 4, 0.2]], 1: [[1, 3, 0.3], [1, 4, 0.4]]}
    a | indexTable(0)
    # returns {3: {0: [[0, 3, 0.1]], 1: [[1, 3, 0.3]]},    4: {0: [[0, 4, 0.2]], 1: [[1, 4, 0.4]]}}
    a | indexTable(1, 0)"""
        self.cols = cols
    def __ror__(self, it):
        if len(self.cols) == 0: return it
        col = self.cols[0]
        return it | groupBy(col) | cli.apply(lambda group: [group[0][col], group | indexTable(*self.cols[1:])]) | cli.toDict()


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""These are pretty standard imports that I always
do. I put it here so that I can do::

    from k1lib.imports import *

and everything will be in place"""
try:
    import torch; from torch import nn, optim
    import torch.nn.functional as F, torch.utils.data as data
except: pass
import matplotlib.pyplot as plt, matplotlib as mpl
import numpy as np, dill as pickle, multiprocessing as mp, concurrent.futures as futures
import math, os, time, sys, random, logging, traceback, re, typing, glob, warnings, asyncio
import dill, json, inspect, xml, base64, io
import functools; from functools import partial, lru_cache
import contextlib; from contextlib import contextmanager
from collections import deque, defaultdict
from typing import List, Tuple, Callable, Union, Iterator, Set, Dict, Any
import k1lib; from k1lib import schedule, graphEqn, mo, knn, fmt, selector,\
viz, Cbs, settings, cli, _k1a, serve, p5, k1ui
from k1lib.cli import *; k1 = k1lib; k1a = _k1a
for e in cli._scatteredClis: globals()[e.__name__] = e
if "py_k1lib_in_applyMp" not in os.environ: k1lib.dontWrap()
plt.rcParams['figure.dpi'] = 100
plt.rcParams["animation.html"] = "jshtml"
from math import e, pi; inf = float("inf"); nan = float("nan"); # this section is for constants
h = 6.62607015e-34; hbar = h/(2*pi); Na = 6.0221408e23; kb = 1.380649e-23
c = 299_792_458; qe = 1.60217663e-19; me = 9.1093837e-31; mn = 1.67262192e-27
e0 = 8.85418782e-12; Dal = u = 1.6605390666e-27; R = 8.3145; sb = 5.670374e-8
if settings.startup.or_patch.numpy: cli.init.patchNumpy()
if settings.startup.or_patch.dict: cli.init.patchDict()
try:
    import pandas as pd
    if settings.startup.or_patch.pandas: cli.init.patchPandas()
except ImportError as e: pass
def dummy():
    """Does nothing. Only here so that you can read source code of this file
and see what's up."""
    pass
try: # these are optional imports
    import torchvision as vision
    import torchvision.datasets as tvDs
    import torchvision.transforms as tf
except: pass
try: import IPython
except: pass
try: import PIL
except: pass
try: from scipy.optimize import fsolve; import scipy
except: pass
try: import einops
except: pass
try: from tqdm import tqdm
except: pass
try: import sympy
except: pass
try: import rdkit; from rdkit import Chem
except: pass
try: import requests
except: pass
try: import seaborn as sns
except: pass


# AUTOGENERATED FILE! PLEASE DON'T EDIT
import re, difflib, k1lib
from typing import Dict, Union, List, Optional
__all__ = ["Eqn", "Eqns", "System"]
settings = k1lib.Settings()
settings.spaceBetweenValueSymbol = True
settings.eqnPrintExtras = True
k1lib.settings.add("eqn", settings, "from k1lib.eqn module");
class Eqn:
    def __init__(self, system:"System"):
        """Creates a blank equation. Not expected to be instantiated by
the end user."""
        self.system = system
        self.terms:Dict[str, float] = {}
    def removeZeros(self):
        self.terms = {k: v for k, v in self.terms.items() if abs(v) > 1e-6}; return self
    def parse(self, line:str):
        line = line.strip().replace(" +", "+").replace("+ ", "+").replace(" ->", "->").replace("-> ", "->")
        reactants, products = line.split("->")
        for i, side in enumerate(line.split("->")):
            sign = i * 2 - 1
            for e in side.split("+"): # side is reactants or products
                e = e.strip()
                number = re.findall("^[0-9.\/]*", e)[0]
                term, number = (e, 1) if number == "" else (e[e.find(number) + len(number):], eval(str(number)))
                term = term.strip(); self.system.terms.add(term)
                if term not in self: self[term] = 0
                self[term] += sign * number
        return self.removeZeros()
    def save(self):
        """Saves this (potentially new) equation to the system, so that it
can be used directly later on"""
        self.system.parse(str(self)); return self
    def __contains__(self, x:str):
        """Whether a term is in this equation"""
        return x in self.terms
    def __getattr__(self, term:str):
        """Gets the value of the term in this equation. Negative if on
consumer side, positive if on producer side"""
        if term in self.terms: return self.terms[term]
        else: return 0
    def __setitem__(self, idx:str, value:float): self.terms[idx] = value; return self
    def __getitem__(self, idx:str):
        """Same as :meth:`__getattr__`"""
        return getattr(self, idx)
    def __iter__(self):
        """Yields key:value pairs"""
        for k, v in self.terms.items(): yield k, v
    def __len__(self):
        """Returns number of terms in this equation"""
        return len(self.terms)
    def __hash__(self): return hash(tuple(self.terms.keys()))
    def __str__(self):
        a = " + ".join((f"{-v}{k}" for k, v in self.terms.items() if v < 0))
        b = " + ".join((f"{v}{k}" for k, v in self.terms.items() if v > 0))
        return f"{a} -> {b}"
    def copy(self):
        answer = Eqn(self.system)
        answer.terms = dict(self.terms); return answer
    def __repr__(self, printExtras=None):
        space = " " if settings.spaceBetweenValueSymbol else ""
        def formatValue(value:float):
            if abs(value - 1) < 1e-9: return ""
            if abs(value - round(value)) < 1e-9:
                return f"{round(value)}{space}"
            return f"{round(value, 3)}{space}"
        a = " + ".join((f"{formatValue(-v)}{k}" for k, v in self.terms.items() if v < 0))
        b = " + ".join((f"{formatValue(v)}{k}" for k, v in self.terms.items() if v > 0))
        answer = f"{a} \033[1m->\033[0m {b}"
        printExtras = printExtras if printExtras is not None else settings.eqnPrintExtras
        return answer if not printExtras else f"""{answer}. Can...
- "MJ" in eqn: to check whether this equation has a specific term
- eqn["MJ"], or eqn.MJ: to get the actual value of the term
- eqn.terms: to get dict of all term -> values
- eqn["MJ"] = 5: to modify a term's value
- eqn * 2: to use normal math operations on the entire equation
- eqn1 @ eqn2: to try to zero out some common terms, useful for unit conversions
- eqn1 == eqn2: see if 2 equations are the same, scale invariant
- for term, value in eqn: to loop over every term and its value
- len(eqn): to get number of terms in the equation
- eqn.copy()"""
    def __mul__(self, number:float):
        answer = self.copy()
        answer.terms = {k: v*number for k, v in self.terms.items()}
        return answer.removeZeros()
    def __rmul__(self, number:float): return self.__mul__(number)
    def __neg__(self): return -1 * self
    def __truediv__(self, number:float):
        answer = self.copy()
        answer.terms = {k: v/number for k, v in self.terms.items()}
        return answer.removeZeros()
    def __rtruediv__(self, number:float): raise Exception("Can't be divided by a number. It doesn't mean anything")
    def __add__(self, eqn):
        answer = self.copy(); answer.terms = {}
        for term, value in self: answer[term] = value + eqn[term]
        for term, value in eqn:
            if term not in answer: answer[term] = value + self[term]
        return answer.removeZeros()
    def __sub__(self, eqn): return self + -1*eqn
    def __eq__(self, eqn):
        if len(self) != len(eqn): return False
        if set(self.terms.keys()) != set(eqn.terms.keys()): return False
        term = list(self.terms.keys())[0]
        eqn = eqn * self[term] / eqn[term]
        for term, value in self:
            if abs(self[term] - eqn[term]) > 1e-9: return False
        return True
    def sharedTerms(self, eqn:"Eqn") -> List[str]:
        """Gets a list of shared terms between this equation and the
specified one."""
        ts = set(self.terms.keys())
        return [t for t in eqn.terms.keys() if t in ts]
    def join(self, eqn:"Eqn", term:str) -> "Eqn":
        """Tries to cancel out this equation with another equation at the
specified term. Example::

    s = eqn.System(\"\"\"a + b -> c + d
    c + 2e -> f\"\"\")
    s.a.c.join(s.c.f, "c") # returns the equation "a + b + 2e -> d + f"

For simpler cases, where the shared term to be joined is obvious, use
:meth:`__matmul__` instead"""
        return self + eqn * (-self[term]/eqn[term])
    def __matmul__(self, eqn:"Eqn") -> "Eqn":
        """Convenience method that does the same thing as :meth:`join`.
Example::

    s = eqn.System(\"\"\"a + b -> c + d
    c + 2e -> f\"\"\")
    s.a.c @ s.c.f # returns the equation "a + b + 2e -> d + f"

Preference order of which term to join:

1) If term is on producer side of ``self``, and consumer side of ``eqn``
2) If term is on consumer side of ``self``, and producer side of ``eqn``
3) Other cases"""
        sharedTerms = self.sharedTerms(eqn)
        def sortF(term):
            if self[term] > 0 and eqn[term] < 0: return 0
            if self[term] < 0 and eqn[term] > 0: return 1
            return 2
        sharedTerms = sorted(sharedTerms, key=sortF)
        if len(sharedTerms) == 0: return None
        return self.join(eqn, sharedTerms[0])
    def round(self, term:str, amount:float=10) -> "Eqn":
        """Rounds the equation off, so that the term's value
is the specified amount. For aesthetic purposes mainly. Example::

    s = eqn.System("a + b -> 2c")
    s.a.c.round("c", 5) # returns the equation "2.5a + 2.5b -> 5c"'"""
        if term not in self: raise AttributeError(term)
        return self * amount / self[term]
    def __round__(self, term:str=None) -> "Eqn":
        """Like :meth:`round`, but more Pythonic?

:param term: Can be any of these:

    - None
    - str
    - Union[int, float]
    - Tuple[str, float]"""
        defaultTerm = list(self.terms.keys())[-1]
        if term is None: return self.round(defaultTerm, 1)
        elif isinstance(term, (tuple, list)): return self.round(*term)
        elif isinstance(term, str): return self.round(term, 1)
        elif k1lib.isNumeric(term): return self.round(defaultTerm, term)
        else: raise AttributeError(f"Don't understand {term}")
class Eqns:
    def __init__(self, system:"System", eqns:List[Eqn], focusTerm:str=None):
        """Creates a new list of equations. Not expected to be instantiated
by the end user.

:param system: injected :class:`System`
:param eqns: list of equations
:param focusTerm: if the list of equations are from the result of focusing
    in a single term, then use this parameter to prioritize certain search
    parameters.
"""
        self.system = system; self.eqns = eqns; self.terms = set()
        for eqn in eqns: self.terms.update(eqn.terms.keys())
        self.focusTerm = focusTerm
    def __getitem__(self, idx:Union[int, str]) -> Optional[Eqn]:
        """If int, return the equation with that index. Not really helpful
for exploring the system of equations, but good for automated scripts

If string, then effectively the same as :meth:`__getattr__`
"""
        return self.eqns[idx] if isinstance(idx, int) else getattr(self, idx)
    def __getattr__(self, term:str) -> Optional[Eqn]:
        """Picks out a specific :class:`Eqn` that has the specified term.
Prefer shorter equations, and the returned :class:`Eqn` always have the
term on the products side. Meaning::

    eqns = eqn.System("a + 2b -> c").b # gets an Eqns object with that single equation
    eqns.a # gets the equation "c -> a + 2b" instead

This is a convenience way to search for equations. If you need more
granularity, use :meth:`pick` instead"""
        chosenEqns = []
        for eqn in self.eqns:
            if term in eqn:
                chosenEqns.append(eqn if eqn[term] > 0 else -eqn)
        chosenEqns = sorted(chosenEqns, key=lambda eqn: len(eqn))
        return None if len(chosenEqns) == 0 else chosenEqns[0]
    def pick(self, *terms:List[str]) -> Optional[Eqn]:
        """Like the quick method (:meth:`__getattr__`), but here, picks
equations more carefully, with selection for multiple terms. Example::

    s = eqn.System(\"\"\"a + 2b -> c
    b + c -> d
    a -> 3d
    a + b + c -> 2d\"\"\")
    s.a.pick("b", "d") # returns last equation
    
As you can see, it's impossible to pick out the last equation using
:meth:`__getattr__` alone, as they will all prefer the shorter equations,
so this is where :meth:`pick` can be useful."""
        chosenEqns = []; t = self.focusTerm or terms[0]
        for eqn in self.eqns:
            if all((term in eqn for term in terms)):
                chosenEqns.append(eqn if eqn[t] > 0 else -eqn)
        chosenEqns = sorted(chosenEqns, key=lambda eqn: len(eqn))
        return None if len(chosenEqns) == 0 else chosenEqns[0]
    def __dir__(self):
        """Returns the list of terms in every equation here. Useful for
tab completion."""
        return list(self.terms)
    def __repr__(self):
        end = """Can...
- eqns[i]: to get the 'i'th equation
- eqns.C: to pick out the first equation that has term 'C'"""
        if self.focusTerm == None:
            eqns = "\n".join([f"{i}. {eqn.__repr__(printExtras=False)}" for i, eqn in enumerate(self.eqns)])
            return f"""Equations:\n{eqns}\n\n{end}"""
        else:
            consumingEqns = []; producingEqns = []
            for eqn in self.eqns:
                if eqn[self.focusTerm] < 0:
                    consumingEqns.append(f"{eqn.__repr__(printExtras=False)}")
                else: producingEqns.append(f"{eqn.__repr__(printExtras=False)}")
            consumingEqns = "\n".join([f"{i}. {eqn}" for i, eqn in enumerate(consumingEqns)])
            producingEqns = "\n".join([f"{i}. {eqn}" for i, eqn in enumerate(producingEqns)])
            return f"""Consumers:\n{consumingEqns}\n\nProducers:\n{producingEqns}\n\n{end}"""
class System:
    def __init__(self, strToParse:str=None):
        """Creates a new system of equations.
        
:param strToParse: if specified, then it gets feed into :meth:`parse`"""
        self.terms = set()
        self.eqns = []
        if strToParse is not None: self.parse(strToParse)
    def parse(self, lines:str) -> "System":
        """Parses extra equations and saves them to this :class:`System`"""
        lines = (line for line in lines.split("\n") if line != "" and not line.startswith("#"))
        self.eqns += [Eqn(self).parse(line) for line in lines if not line.startswith("#")]
        self.eqns = list(set(self.eqns))
        return self
    def spellCheck(self):
        """Runs a spell check to find out terms that are pretty similar to
each other"""
        print("Similar terms:"); terms = list(self.terms)
        for i, iTerm in enumerate(terms):
            for j, jTerm in enumerate(terms[i+1:]):
                if iTerm[:-1] == jTerm[:-1]: continue
                if abs(len(iTerm) - len(jTerm)) > 2: continue
                r = difflib.SequenceMatcher(None, iTerm, jTerm).ratio()
                if r < 0.9: continue
                print(f"- {round(r*100)}% similar: {iTerm}, {jTerm}")
    def __len__(self): return len(self.eqns)
    def __getitem__(self, idx:int) -> Eqn:
        """Picks out the i'th equation from the list of equations. Useful
for automated scripts"""
        return self.eqns[idx]
    def __getattr__(self, term:str) -> Eqns:
        """Picks out equations that has the term"""
        return Eqns(self, [eqn for i, eqn in enumerate(self.eqns) if term in eqn], focusTerm=term)
    def __dir__(self):
        """Returns the list of terms in every equation here. Useful for
tab completion."""
        return list(self.terms)
    def __repr__(self):
        return f"""System of {len(self)} equations:\n{Eqns(self, self.eqns)}\n
Can...
- s[i]: to get a specific equation
- s.C: to get all equations that involve a specific substance "C"
- s.spellCheck(): to check if there are terms that are close to each other
"""


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""This is for everything related to ipython notebooks. Expected to use behind the
"nb" module name, like this::

    from k1lib.imports import *
    nb.execute("file.ipynb")"""
__all__ = ["cells", "pretty", "execute"]
from k1lib.cli import BaseCli; import k1lib.cli as cli
import json, k1lib, os, traceback; from typing import List
import matplotlib.pyplot as plt
def _cells(fileName, outputs=False):
    js = json.loads(cli.cat(fileName) | cli.join("\n"))
    cells = []; fields = set(["cell_type", "source"])
    if outputs: fields.add("outputs")
    return [{k:cell[k] for k in cell.keys() if k in fields} for cell in js["cells"]]
def cells(fileName=None, outputs=False):
    """Gets simplified notebook cells from file source, including fields
``cell_type`` and ``source`` only. Example::

    nb.cells("file.ipynb")"""
    if fileName is None: return cli.aS(_cells, outputs)
    else: return _cells(fileName, outputs)
class pretty(BaseCli):
    def __init__(self, magics:bool=False, whitelist:List[str]=[], blacklist:List[str]=[]):
        """Makes the cells prettier.
Cell 1 in file.ipynb::

    #notest, export
    a = 3

Cell 2 in file.ipynb::

    b = 6

Code::

    # only cell 2 gets chosen
    nb.cells("file.ipynb") | nb.pretty(blacklist=["notest"])
    # only cell 1 gets chosen
    nb.cells("file.ipynb") | nb.pretty(whitelist=["export"])

:param magics: if False, then if detected magics ('!', '%%' symbols), then remove that
    line in cell's source
:param whitelist: every cell that doesn't have any of these properties will be
    filtered out
:param blacklist: every cell that has any of these properties will be filtered out"""
        self.magics = magics; self.wl = whitelist; self.bl = blacklist
    def __ror__(self, cells):
        magics = self.magics; wl = self.wl; bl = self.bl
        for cell in cells:
            lines = cell["source"]
            if not magics: lines = [line.rstrip() for line in lines if not line.startswith("!") and not line.startswith("%%")]
            if len(lines) == 0: continue
            if lines[0].startswith("#"): props = set([e.strip() for e in lines[0].lstrip("#").split(",")])
            else: props = set()
            if len(wl) > 0 and not any([e in props for e in wl]): continue
            if len(bl) > 0 and any([e in props for e in bl]): continue
            cell["source"] = lines; yield cell
class execute(BaseCli):
    def __init__(self, fileName=None, _globals:dict=None):
        """Executes cells.
Example::

    nb.cells("file.ipynb") | nb.execute("nb.ipynb")

Most of the time, you'd want to pass cells through :class:`pretty` first, to make sure
everything is nice and clean

:param fileName: not actually used to read the file. If specified, then changes the
    current working directory to that of the file
:param _globals: optional dict of global variables"""
        self.fileName = fileName
        self._globals = dict() if _globals is None else _globals
    def __ror__(self, cells):
        with k1lib.settings.context():
            if self.fileName:
                k1lib.settings.wd = os.path.dirname(self.fileName) or "."
            for cell in cells:
                if cell["cell_type"] != "code": continue
                source = "\n".join(cell["source"] | cli.apply(lambda x: x.strip("\n")))
                try: exec(source, self._globals); plt.show()
                except Exception as e:
                    print("Problematic cell:\n")
                    print(source); traceback.print_exc(); raise e


# AUTOGENERATED FILE! PLEASE DON'T EDIT
from .callbacks import Callback, Callbacks, Cbs
import k1lib; from k1lib import squeeze; import k1lib.cli as cli
from functools import partial
import matplotlib.pyplot as plt
from typing import List, Tuple, Dict, Iterator, Union, Any, Callable
try: import torch; import torch.nn as nn; hasTorch = True
except:
    torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))
    nn = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["HookModule"]
class Handles:
    def __init__(self):
        self.forward = None; self.backward = None
    def remove(self):
        if self.active:
            self.forward.remove(); self.forward = None
            self.backward.remove(); self.backward = None
    @property
    def active(self):
        if self.forward != None and self.backward != None: return True
        elif self.forward == None and self.backward == None: return False
        raise Exception("Supposed to be unreachable")
class Data(k1lib.Object):
    def __init__(self):
        super().__init__(); self.withAutoDeclare(lambda: [])
class ModuleData:
    def __init__(self): self.forward = Data(); self.backward = Data()
    def _plot(self, axes, field:str, rangeSlice:slice, f):
        forwardData = self.forward[field]; step = rangeSlice.step or 1
        backwardData = self.backward[field]
        if len(forwardData) == 0 or len(backwardData) == 0: return
        fR, bR = k1lib.Range.proportionalSlice(len(forwardData), len(backwardData), rangeSlice)
        axes[0].plot(fR.range_[::step] | f | cli.deref(), forwardData[fR.slice_][::step] | f | cli.deref(), alpha=0.5)
        axes[1].plot(bR.range_[::step] | f | cli.deref(), backwardData[bR.slice_][::step] | f | cli.deref(), alpha=0.5)
    def __repr__(self):
        return """Module's saved data. can...
- d.forward: to get data stored during forward pass
- d.backward: to get data stored during backward pass"""
_Fn = Callable[[Data, nn.Module, Tuple[torch.Tensor], Tuple[torch.Tensor]], None]
class Function:
    def __init__(self, f:_Fn, name=None):
        self.f = f; self.name = name or "f(<no name>)"
    def __call__(self, *args, **kwargs):
        self.f(*args, **kwargs)
def hook(every, fns:List[Function], *args):
    if every.value: [fn(*args) for fn in fns]
class Module:
    def __init__(self, module:nn.Module):
        self.nn = module
        self.handles = Handles()
        self.data = ModuleData()
        self.name = module.__class__.__name__
    def registerHooks(self, forwardFns:List[Function], backwardFns:List[Function], every):
        self.handles.forward = self.nn.register_forward_hook(partial(hook, every, forwardFns, self.data.forward))
        self.handles.backward = self.nn.register_full_backward_hook(partial(hook, every, backwardFns, self.data.backward))
        return self
    def unregisterHooks(self): self.handles.remove()
    def __repr__(self):
        return f"""Module `{self.name}`. Use...
- m.data: to get data stored
- m.nn: to get actual nn.Module object
- m.plot("means", "stds"): to plot simple statistics"""
@k1lib.patch(Cbs)
class HookModule(Callback):
    """Hooks into selected modules in the network, and
execute functions like .mean(), .std(). This is fairly
complicated, and I highly recommend displaying this
callback in a cell for more info"""
    def __init__(self, persistent:bool=False):
        """
:param persistent: whether to save results across
    runs. If false, then can execute `.reset()` to
    reset everything"""
        super(HookModule, self).__init__()
        self.modules:List[Module] = []
        self.forwardFns:List[Function] = []
        self.backwardFns:List[Function] = []
        self.cleanFns = []; self.persistent = persistent
        self.every = k1lib.Every(3)
    def reset(self):
        """Intended to be called by end user only, to reset
        everything if choose to persist results across runs."""
        self._end(); self._start()
    def persist(self):
        """By default, data will be erased and populated on each run. If
you want the data to persist across runs, call this."""
        self.persistent = True
    def startRun(self):
        if (not self.persistent) or (len(self.modules) == 0): self._start()
    def startBatch(self): self.every()
    def _registerHooks(self):
        for module in self.modules:
            module.registerHooks(self.forwardFns, self.backwardFns, self.every)
    def _unregisterHooks(self):
        for module in self.modules: module.unregisterHooks()
    def endRun(self):
        if not self.persistent: self._end()
    def suspend(self):
        self.actuallyRestore = len(self) == 0 or self[0].handles.active
        if self.actuallyRestore: self._unregisterHooks()
    def restore(self):
        if self.actuallyRestore:
            self._registerHooks()
            self.actuallyRestore = False
    def __getitem__(self, idx):
        if type(idx) == int: return self.modules[idx]
        answer = HookModule(self.persistent)
        answer.modules = self.modules[idx]
        return answer
    def __len__(self): return len(self.modules)
    def __repr__(self):
        f = '\n'.join([f'  - {fn.name or str(fn)}' for fn in self.forwardFns])
        f = "" if f == "" else f"Forward hooks:\n{f}\n"
        b = '\n'.join([f'  - {fn.name or str(fn)}' for fn in self.backwardFns])
        b = "" if b == "" else f"Backward hooks:\n{b}\n"
        n = '\n'.join([f'  {i}. {data.name}' for i, data in enumerate(self)])
        excludes = {"withForwardHook", "withBackwardHook", "withHook", "withCheckpoint"}
        withs = '\n'.join([f"- m.{key}()" for key in dir(self) if key.startswith("with") and key not in excludes])
        return f"""{super()._reprHead} with {len(self)} modules:\n{n}\n{f}{b}
Use...
- m.plot("means", "stds"): to plot simple statistics
- m[i]: to get a specific module
- m[a:b]: to get a new HookModule with selected modules
- m.css("..."): to select a specific subset of modules only
- m.withHook(hookCb): to hook a specific callback function
- m.clearHooks(): to clear all hooks
{super()._reprCan}

Built-in `with-` functions:\n{withs}"""
@k1lib.patch(HookModule)
def _start(self):
    self.modules = []
    for nn, sel in zip(self.l.model.modules(), self.l.selector.modules()):
        if "HookModule" in sel: self.modules.append(Module(nn))
    self._registerHooks()
@k1lib.patch(HookModule)
def _end(self):
    for module in self.modules:
        for cleanFn in self.cleanFns:
            cleanFn(module.data)
    self._unregisterHooks()
@k1lib.patch(HookModule)
def withForwardHook(self, hook:_Fn, name:str=None):
    """Adds a hook to the forward pass. See :func:`~k1lib.callbacks.hookModule.HookModule.withHook`"""
    self.forwardFns += [Function(hook, name)]; return self
@k1lib.patch(HookModule)
def withBackwardHook(self, hook:_Fn, name:str=None):
    """Adds a hook to the backward pass. See :func:`~k1lib.callbacks.hookModule.HookModule.withHook`"""
    self.backwardFns += [Function(hook, name)]; return self
@k1lib.patch(HookModule)
def withHook(self, hook:_Fn, name:str=None):
    """Adds a hook to both the forward and backward pass.

:param hook: this function is expected to take in these parameters: **(data, module, inp, out)**

    :data: the injected dependency for you to store stuff.
        Initially, `data.max` is an empty list, and you can
        append to it directly, like this::

            data.max.append() # okay

        Later on, you can do things like::

            HookModule[i].forward.max

        and get the data you saved from the hook.
    :module: the module this function hooks into. Please
        refer to :func:`torch.nn.Module.register_forward_hook()` to
        know more.
    :inp: input (or grad of input) to the module
    :out: output (or grad of output) to the module
:param name: custom name for the function for nice displaying

See also: m.withForwardHook(), m.withBackwardHook()"""
    return self.withForwardHook(hook, name).withBackwardHook(hook, name)
@k1lib.patch(HookModule)
def clearHooks(self):
    self._unregisterHooks()
    self.forwardFns = []; self.backwardFns = []
    self.cleanFns = []; return self
def meanCb(data, m, inp, out):
    data.means.append(squeeze(out, hard=True).data.mean().item())
@k1lib.patch(HookModule)
def withMeanRecorder(self):
    """Records mean"""
    return self.withHook(meanCb, "mean")
def stdCb(data, m, inp, out):
    data.stds.append(squeeze(out, hard=True).data.std().item())
@k1lib.patch(HookModule)
def withStdRecorder(self):
    """Records standard deviation"""
    return self.withHook(stdCb, "std")
def minCb(data, m, inp, out):
    data.mins.append(squeeze(out, hard=True).data.min().item())
@k1lib.patch(HookModule)
def withMinRecorder(self):
    """Records min"""
    return self.withHook(minCb, "min")
def maxCb(data, m, inp, out):
    data.maxs.append(squeeze(out, hard=True).data.max().item())
@k1lib.patch(HookModule)
def withMaxRecorder(self):
    """Records max"""
    return self.withHook(maxCb, "max")
@k1lib.patch(HookModule)
def css(self, css:str):
    answer = HookModule()
    selector = k1lib.selector.select(self.l.model, css)
    d = {m.nn: m for m in self.modules}
    for sel in selector.modules():
        if "HookModule" in sel and sel.nn in d:
            answer.modules.append(d[sel.nn])
    return answer
def plotF(modules:HookModule, fields:List[str], f, rangeSlice:slice):
    fig, axes = plt.subplots(len(fields), 2, figsize=(10, 3*len(fields)), dpi=100)
    axes = axes.reshape((-1, 2))
    for axs, field in zip(axes, fields):
        for module in modules:
            module.data._plot(axs, field, rangeSlice, f)
        axs[0].set_title(f"Forward {field}")
        axs[1].set_title(f"Backward {field}")
    plt.figlegend([f"{i}. {module.name}" for i, module in enumerate(modules)], loc='center right')
@k1lib.patch(HookModule)
@k1lib.patch(Module)
def plot(self, *fields:List[str], f=cli.iden()):
    """Plots every simple (1 number saved/pass/module) fields.

:param fields: list of fields to plot. If none, then
    will automatically find all simple fields"""
    modules = [self] if isinstance(self, Module) else self
    if len(modules) == 0: raise Exception("No modules to plot!")
    if len(fields) == 0:
        fields = []; forwardData = modules[0].data.forward
        for field in forwardData.state.keys():
            if field.startswith("_"): continue
            fieldData = forwardData[field]
            if type(fieldData) == list and k1lib.isNumeric(fieldData[0]):
                fields.append(field)
    return k1lib.viz.SliceablePlot(partial(plotF, modules, fields, f))


# AUTOGENERATED FILE! PLEASE DON'T EDIT
from .callbacks import Callback, Callbacks, Cbs
import k1lib
import matplotlib.pyplot as plt
from functools import partial
from typing import List, Tuple, Callable, Union
try: import torch; import torch.nn as nn; hasTorch = True
except:
    torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))
    nn = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["HookParam"]
class ParamData(k1lib.Object):
    def __init__(self):
        super().__init__()
        self.means = []; self.stds = []
        self.mins = []; self.maxs = []
    def update(self, torchParam:nn.Parameter):
        self.means.append(torchParam.mean().item())
        self.stds.append(torchParam.std().item())
        self.mins.append(torchParam.min().item())
        self.maxs.append(torchParam.max().item())
    def __len__(self): return len(self.means)
    def __repr__(self):
        return f"""Param's saved data. Use...
- d.means: to get list of means
- d.stds: to get list of means
- d.mins: to get list of mins
- d.maxs: to get list of maxs"""
class Param:
    def __init__(self, name:str, torchParam:nn.Parameter):
        self.name = name
        self.torchParam = torchParam
        self.data = ParamData()
        self.every = k1lib.Every(3)
    def update(self):
        if self.every(): self.data.update(self.torchParam.detach())
    def __repr__(self):
        return f"""Param `{self.name}`. Use...
- p.torchParam: to get actual underlying parameter
- p.data: to get data stored
- cb.plot(): to quickly look at everything"""
@k1lib.patch(Cbs)
class HookParam(Callback):
    """Records means and stds of all parameters"""
    def __init__(self):
        ""
        super().__init__(); self.params:List[Param] = []
    def __getitem__(self, idx:Union[int, slice]):
        if type(idx) == int: return self.params[idx]
        answer = HookParam(); answer.params = self.params[idx]; return answer
    def __len__(self): return len(self.params)
    def _selected(self, paramName:str):
        splits = paramName.split(".")
        try:
            mS = self.l.selector
            for split in splits[:-1]: mS = mS[split]
            return "HookParam" in mS and hasattr(mS, splits[-1])
        except KeyError: return False
    def startRun(self):
        if len(self) == 0: # set things up first time only
            self.params = [Param(k, v) for k, v in self.l.model.named_parameters() if self._selected(k)]
    def startBatch(self): [param.update() for param in self.params]
    def css(self, css:str):
        """Creates a new HookParam object with selected modules. May be useful
for displaying a subset of the recorded data"""
        oldSelector = self.l.selector; answer = HookParam()
        self.l.selector = k1lib.selector.select(self.l.model, css)
        answer.params = [param for param in self.params if self._selected(param.name)]
        self.l.selector = oldSelector; return answer
    def __repr__(self):
        s = f", {len(self[0].data)} means and stds each" if len(self) > 0 else ""
        names = "\n".join([f"  {i}. {p.name}" for i, p in enumerate(self)])
        return f"""{super()._reprHead}: {len(self)} params{s}:\n{names}\n
Use...
- p.plot(): to quickly look at everything
- p[i]: to view a single param
- p[a:b]: to get a new HookParam with selected params
- p.css("..."): to select a specific subset of modules only
{super()._reprCan}"""
def plotF(params:Union[HookParam, Param, List[Param]], rangeSlice:slice):
    if type(params) == Param: params = [params]
    fields = params[0].data.state.keys(); step = rangeSlice.step or 1
    fig, axes = plt.subplots(2, 2, figsize=(10, 6), dpi=100)
    axes = axes.flatten()
    for field, ax in zip(fields, axes):
        for param in params:
            fieldData = param.data[field]
            r = k1lib.Range(len(fieldData))[rangeSlice]
            ax.plot(r.range_[::step], fieldData[r.slice_][::step])
        ax.set_title(field.capitalize())
    plt.figlegend([p.name for p in params], loc='right')
@k1lib.patch(HookParam)
@k1lib.patch(Param)
def plot(self): return k1lib.viz.SliceablePlot(partial(plotF, self))


# AUTOGENERATED FILE! PLEASE DON'T EDIT
from .callbacks import Callback, Callbacks, Cbs
import k1lib, time, math, logging, numpy as np
from functools import partial
import matplotlib.pyplot as plt
try: import torch; from torch import nn; hasTorch = True
except: hasTorch = False
__all__ = ["Profiler"]
if hasTorch:
    import k1lib.callbacks.profilers as ps
    ComputationProfiler = ps.computation.ComputationProfiler
    IOProfiler = ps.io.IOProfiler
    MemoryProfiler = ps.memory.MemoryProfiler
    TimeProfiler = ps.time.TimeProfiler
else:
    class ComputationProfiler: pass
    class IOProfiler: pass
    class MemoryProfiler: pass
    class TimeProfiler: pass
@k1lib.patch(Cbs)
class Profiler(Callback):
    """Profiles memory, time, and computational complexity of the network. See over
:mod:`k1lib.callbacks.profilers` for more details on each of these profilers"""
    def __init__(self):
        super().__init__(); self.clear(); self.dependsOn=["Recorder"]
    def clear(self):
        """Clears every child profilers"""
        self._mpCache=None; self._tpCache=None
        self._cpCache=None; self._ioCache=None
    def _memory(self): # do this to quickly debug, cause if not, Callback will just raise AttributeError on .memory
        if self._mpCache != None: return self._mpCache
        with self.cbs.context():
            mp = MemoryProfiler(); self.cbs.add(mp)
            mp._run(); self._mpCache = mp; return mp
    @property
    def memory(self) -> MemoryProfiler:
        """Gets :class:`~k1lib.callbacks.profilers.memory.MemoryProfiler`"""
        return self._memory()
    def _computation(self):
        if self._cpCache != None: return self._cpCache
        with self.cbs.context():
            cp = ComputationProfiler(self); self.cbs.add(cp)
            cp._run(); self._cpCache = cp; return cp
    @property
    def computation(self) -> ComputationProfiler:
        """Gets :class:`~k1lib.callbacks.profilers.computation.ComputationProfiler`"""
        return self._computation()
    def _time(self):
        if self._tpCache != None: return self._tpCache
        with self.cbs.context():
            tp = TimeProfiler(); self.cbs.add(tp)
            tp._run(); self._tpCache = tp; return tp
    @property
    def time(self) -> TimeProfiler:
        """Gets :class:`~k1lib.callbacks.profilers.time.TimeProfiler`"""
        return self._time()
    def _io(self):
        if self._ioCache != None: return self._ioCache
        with self.cbs.context():
            io = IOProfiler(); self.cbs.add(io)
            io._run(); self._ioCache = io; return io
    @property
    def io(self) -> IOProfiler:
        """Gets :class:`~k1lib.callbacks.profilers.io.IOProfiler`"""
        return self._io()
    def __repr__(self):
        return f"""{self._reprHead}, can...
- p.memory: to profile module memory requirements
- p.time: to profile module execution times
- p.computation: to estimate module computation
- p.io: to get input and output shapes of 
{self._reprCan}"""


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""This module is for creating dynamic graphs using plain old
equations. For example::

    from k1lib.imports import *
    x = graphEqn.Variable()
    y = x * 3 + 5
    z = y ** 5
    
    z(2) # returns 161051 (from (2 * 3 + 5) ** 5)

Point is, ``x`` is an unknown, ``y`` is a "function" of ``x``. ``z`` depends on
``y``, but is also a function of ``x``.

Remember that you can go pretty wild with this::

    x2 = k1lib.inverse(z)
    x2(161051) # returns 2.0

Here, ``x2`` is actually a function x(z).

For simple functions like this, it should take 200us to solve it. You can also
declare a bunch of variables early on, and then resolve them one by one like
this::

    a = Variable(); b = Variable()
    c = a + b + 2; a.value = 6
    c(5) # returns 13

    b.value = 7
    c() # returns 15
"""
from typing import Callable as _Callable, Union as _Union, Iterator as _Iterator
import k1lib as _k1lib
__all__ = ["Variable"]
F = _Callable[[float, float], float]
class Expression:
    def __init__(self, a:"Variable", b:"Variable", operation:F):
        self.a = a; self.b = b; self.operation = operation
    @property
    def resolved(self):
        """Whether this expression has been resolved (both internal variables are
resolved)."""
        return self.a.resolved and self.b.resolved
    @property
    def value(self):
        """Value of the expression."""
        return self.operation(self.a._value, self.b._value)
    def applyF(self, f:_Callable[["Variable"], None]):
        self.a._applyF(f); self.b._applyF(f)
def _op2(a, b, operation:F):
    a = a if isinstance(a, Variable) else _Constant(a)
    b = b if isinstance(b, Variable) else _Constant(b)
    answer = Variable(); answer.expr = Expression(a, b, operation)
    if answer.expr.resolved: answer._value = answer.expr.value
    return answer
class Variable:
    _idx = 0
    def __init__(self):
        self.__class__._idx += 1; self.variableName = f"V{self.__class__._idx}"
        self.expr:Expression = None
        self._value:float = None # if not None, then already resolved
        self.isConstant = False # to know if the value above is resolved, or is truely a literal number
        self.trial:int = 0 # current resolve trial number
    @property
    def value(self) -> _Union[float, None]:
        """Actual float value of :class:`Variable`. When setting this, if the
new value's not None, the object would act like a constant in every future
equations. To turn it back into a :class:`Variable`, simply set this to
:data:`None`."""
        return self._value
    @value.setter
    def value(self, v):
        """Sets the value of variable. If it's an actual value, """
        if v is None: self._value = None; self.isConstant = False
        else: self._value = v; self.isConstant = True
    def _reset(self): self._value = self._value if self.isConstant else None
    @property
    def resolved(self):
        """Whether this variable has been resolved or not."""
        return self._value != None
    def _applyF(self, f:_Callable[["Variable"], None]): # apply an operation to variable and its dependencies
        f(self)
        if self.expr != None: self.expr.applyF(f)
    @property
    def _leaves(self) -> _Iterator["Variable"]:
        """Get variables that does not have an expression linked to it. Aka at
the leaf."""
        if self.resolved: return
        if self.expr == None: yield self
        else:
            yield from self.expr.a._leaves
            yield from self.expr.b._leaves
    @property
    def leaves(self): return list(set(self._leaves))
    def __call__(self, x:float=None) -> _Union[float, None]:
        """Tries to solve this variable given the independent variable ``x``.

:param x: if nothing is specified, you have to be sure that all variables already
    have a value."""
        return self._solve(x)
    def __add__(self, v): return _op2(self, v, lambda a, b: a + b)
    def __sub__(self, v): return _op2(self, v, lambda a, b: a - b)
    def __neg__(self): return _op2(_Constant(0), self, lambda a, b: a - b)
    def __mul__(self, v): return _op2(self, v, lambda a, b: a * b)
    def __truediv__(self, v): return _op2(self, v, lambda a, b: a / b)
    def __pow__(self, v): return _op2(self, v, lambda a, b: a**b)
    def __radd__(self, v): return _op2(v, self, lambda a, b: a + b)
    def __rsub__(self, v): return _op2(v, self, lambda a, b: a - b)
    def __rmul__(self, v): return _op2(v, self, lambda a, b: a * b)
    def __rtruediv__(self, v): return _op2(v, self, lambda a, b: a / b)
    def __rpow__(self, v): return _op2(v, self, lambda a, b: a**b)
    def __repr__(self): return f"{self._value}" if self.resolved else f"<Variable {self.variableName}>"
    def __int__(self): return self._value
    def __float__(self): return self._value
@_k1lib.patch(Variable)
def _resolve(self, trial:int) -> bool:
    """Attempts to resolve variable. Return true if expression tree under
this Variable changes at all.

:param trial: how many times _resolve() has been called by the originating
    :class:`Variable`? Only updates stuff if in a new trial."""
    if self.trial >= trial or self.resolved or self.expr == None: return False
    # try to resolve dependencies first
    changed = self.expr.a._resolve(trial) or self.expr.b._resolve(trial)
    self.trial = trial
    if self.expr.resolved: self._value = self.expr.value; changed = True
    return changed
@_k1lib.patch(Variable)
def _simplify(self, printStuff:bool=False):
    """Simplify system before solving"""
    self._applyF(lambda v: setattr(v, "trial", 0)); trial = 2
    while self._resolve(trial): trial += 1
    if printStuff and not self.resolved: print("Can't find a solution")
@_k1lib.patch(Variable)
def _solve(self, x:float) -> _Union[float, None]:
    """Try to solve this expression tree, given value of independent
variable."""
    self._applyF(lambda v: v._reset()); self._simplify(); leaves = self.leaves
    if len(leaves) > 1: raise Exception(f"System of equation has {len(leaves)} indenpendent variables. Please constrain system more!")
    elif len(leaves) == 1: next(iter(leaves))._value = x
    self._simplify(True); return self._value
class _Constant(Variable):
    def __init__(self, value:float):
        """Creates a constant :class:`Variable` with some specified value."""
        super().__init__(); self._value = value; self.isConstant = True


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""This is for core callbacks, that defines how everything is going to go"""
from .callbacks import Callback, Callbacks, Cbs
import k1lib; from typing import List, Tuple, Dict, Iterator, Union, Any, Callable
try: import torch; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["CoreNormal", "CoreRNN"]
@k1lib.patch(Cbs)
class CoreNormal(Callback):
    """Just a normal, typical feed forward pass.
Deposits variables into :class:`~k1lib.Learner` at checkpoint ``inPass``:

- y: attached result tensor after passing through model"""
    def inPass(self):
        self.l.y = self.l.model(self.l.xb)
@k1lib.patch(Cbs)
class CoreRNN(Callback):
    """RNN forward pass.
Expected variables from :attr:`k1lib.Learner.model`:

- initHidden: function takes in batch size, returns init hidden tensor

Deposits variables into :class:`~k1lib.Learner` at checkpoint ``inPass``, more
specifically ``rnnPass``:

- y: attached result tensor after pass (``inPass``), after character pass (``rnnPass``)
"""
    def startBatch(self):
        self.hx = self.l.model.initHidden(self.l.xb.shape[-2])
    def inPass(self):
        self.hx = self.hx.to(self.l.xb.device)
        for item in self.l.xb:
            self.l.y, self.hx = self.l.model(item, self.hx)
            self.cbs("rnnPass")


# AUTOGENERATED FILE! PLEASE DON'T EDIT
from .atom import Atom, mo, _a
from functools import partial
from typing import List
import k1lib, re
__all__ = ["alkane", "alcohol", "sameEmpirical"]
def _mole(name, f):
    _a[name] = f; mo.registerSubstance(name, f)
_mole("H2O", lambda: mo.O(mo.H)(mo.H))
_mole("CH4", lambda: mo.C(mo.H)(mo.H)(mo.H)(mo.H))
_mole("COOH", lambda: mo.C(mo.O, 2)(mo.O(mo.H))(mo.H))
_mole("NH3", lambda: mo.N(mo.H)(mo.H)(mo.H))
_mole("CH3OH", lambda: mo.CH4(mo.H2O))
def glucose():
    o = mo.O
    o.bond(mo.CH4).bond(mo.CH3OH).bond(mo.CH3OH).bond(mo.CH3OH).bond(mo.CH3OH).bond(o)
    o.next().bond(mo.CH3OH); return o
_mole("glucose", glucose)
def cyclohexane():
    c = mo.CH4
    return c.bond(mo.CH4).bond(mo.CH4).bond(mo.CH4).bond(mo.CH4).bond(mo.CH4).bond(c)
_mole("cyclohexane", cyclohexane)
def benzene():
    c = mo.CH4
    return c.main(mo.CH4, 2).main(mo.CH4).main(mo.CH4, 2).main(mo.CH4).main(mo.CH4, 2).main(c)
_mole("benzene", benzene)
def adenine():
    n = mo.NH3
    c1 = n.bond(mo.CH4).bond(mo.NH3, 2).bond(mo.CH4)
    c2 = c1.bond(mo.CH4, 2); c2.bond(n)
    c1.bond(mo.CH4)(mo.NH3).bond(mo.NH3, 2).bond(mo.CH4).bond(mo.NH3, 2).bond(c2);
    return n
_mole("adenine", adenine)
def ribose():
    o = mo.H2O
    return o.bond(mo.CH4(mo.CH3OH)).main(mo.CH3OH).main(mo.CH3OH).main(mo.CH3OH).main(o)
_mole("ribose", ribose)
def adenosine(): ri = mo.ribose; ri.next(1)(mo.adenine); return ri
_mole("adenosine", adenosine)
def alkane(n:int) -> Atom:
    """Creates an alkane with ``n`` carbons.
Example::

    # returns "C3H8"
    mo.alkane(3).empirical()"""
    answer = mo.CH4
    for i in range(n-1): answer = answer.bond(mo.CH4)
    return answer
def alcohol(n:int, loc:int=1) -> Atom:
    """Creates an alcohol with ``n`` carbons and an OH group at carbon ``loc``.
Example::

    # returns "C3H8O"
    mo.alcohol(3, 1).empirical()"""
    a = alkane(n)
    a.next(times=loc-1)(mo.H2O)
    return a
@k1lib.patch(Atom)
def perfluoro_ize(mainA:Atom) -> Atom:
    """Replaces all C-H bonds with C-F bonds"""
    for a in mainA.atoms():
        b = a.next()
        if a.name == "H" and b.name == "C":
            a.removeBond(b); b(mo.F)
    return mainA
empC = re.compile("[A-Z][a-z]*[0-9]*")
def sameEmpirical(a:str, b:str) -> bool:
    """Checks whether 2 empirical formula are the same or not.
Example::

    moparse.sameEmpirical("C2H4", "H4C2") # returns True"""
    return sorted(re.findall(empC, a)) == sorted(re.findall(empC, b))


import inspect, io, base64
from flask import Flask, request
from flask_cors import CORS

d = k1lib.serve.analyze(endpoint); args = d["args"]; annos = d["annos"]
d | aS(dill.dumps) | file("META_FILE")
app = Flask(__name__); CORS(app)

@app.route("/healthCheck", methods=["GET", "POST"])
def healthCheck(): return "ok"

@app.route("/", methods=["POST"])
def main():
    js = request.json
    return k1lib.serve.pyToWeb(endpoint(*[k1lib.serve.webToPy(js[arg], annos[arg]) for arg in args]), annos["return"])

try:
    import waitress
    waitress.serve(app, host='0.0.0.0', port=SOCKET_PORT)
except: app.run(host='0.0.0.0', port=SOCKET_PORT)






# AUTOGENERATED FILE! PLEASE DON'T EDIT
import io, sys, warnings, time, k1lib, os
from contextlib import contextmanager
from functools import partial
try: import rdkit; hasRdkit = True
except: hasRdkit = False
__all__ = ["captureStdout", "ignoreWarnings", "timer", "attrContext"]
@contextmanager
def captureStdout(out=True, c=False) -> k1lib.Wrapper:
    """Captures every print() statement. Taken from https://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call.
Example::

    with k1lib.captureStdout() as outer:
        print("something")
        with k1lib.captureStdout() as inner:
            print("inside inner")
        print("else")
    # prints "['something', 'else']"
    print(outer.value)
    # prints "['inside inner']"
    print(inner.value)

Note that internally, this replaces :data:`sys.stdout` as :class:`io.StringIO`, so
might not work property if you have fancy :class:`bytes` stuff going on. Also,
carriage return (``\\r``) will erase the line, so multi-line overlaps might not
show up correctly.

If you wish to still print stuff out, you can do something
like this::

    with k1.captureStdout() as out:
        print("abc") # gets captured
        out.print("def") # not captured, will actually print out
        # you can also access the stream like this
        print("something", file=out.old)

Also, by default, this won't work if you're trying to capture C
library's output, because they write to stdout directly, instead of
going through Python's mechanism. You can capture it down at the C
level by doing this::

    with captureStdout() as out1, captureStdout(c=True) as out2:
        os.system("ls") # gets captured by out2, but not out1
        print("abc")    # gets captured by out1, but not out2

It's a little bit hard to actually integrate C mode and non-C mode
together, so for the time being, you gotta have 2 context managers
if you want to capture absolutely everything, C or not.

:param out: if True, captures stdout, else captures stderr
:param c: whether to capture at the C/C++ level or not"""
    if c is True:
        import wurlitzer; w = k1lib.Wrapper("")
        try:
            with wurlitzer.pipes() as (_out, _err): yield w
        finally: w.value = _out.read().split("\n") if out else _err.read().split("\n")
    else:
        if out: _stdout = sys.stdout; sys.stdout = _stringio = io.StringIO()
        else:   _stdout = sys.stderr; sys.stderr = _stringio = io.StringIO()
        w = k1lib.Wrapper([])
        w.print = partial(print, file=_stdout)
        w.old = _stdout
        try: yield w
        finally:
            w.value = [l.split("\r")[-1] for l in _stringio.getvalue().split("\n")]
            if out: sys.stdout = _stdout
            else:   sys.stderr = _stdout
@contextmanager
def ignoreWarnings():
    """Context manager to ignore every warning.
Example::

    import warnings
    with k1lib.ignoreWarnings():
        warnings.warn("some random stuff") # will not show anything"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        warnings.filterwarnings('ignore')
        if hasRdkit: rdkit.RDLogger.DisableLog('rdApp.*')
        yield
    if hasRdkit: rdkit.RDLogger.EnableLog('rdApp.*')
@contextmanager
def timer():
    """Times generic code.
Example::

    with k1lib.timer() as t:
        time.sleep(1.1)
    # prints out float close to 1.1
    print(t())

The with- statement will actually return a :class:`~k1lib.Wrapper` with value
None. The correct time will be deposited into it after the code block ends."""
    w = k1lib.Wrapper(None)
    beginTime = time.time()
    try: yield w
    finally: w.value = time.time() - beginTime
@contextmanager
def attrContext(var, **kwargs):
    """Temporarily sets variable's attribute to something else.
Example::

    class A: pass
    a = A()
    a.b = 3
    print(a.b) # prints "3"
    with k1lib.attrContext(a, b=4, c=5):
        print(a.b, a.c) # prints "4 5"
    print(a.b, a.c) # prints "3 None"
"""
    oldValues = dict()
    for k, v in kwargs.items():
        oldValues[k] = getattr(var, k, None)
        setattr(var, k, v)
    try: yield
    finally:
        for k, v in oldValues.items():
            setattr(var, k, v)


# AUTOGENERATED FILE! PLEASE DON'T EDIT
from k1lib.callbacks import Callback, Cbs; from k1lib import fmt, cli
import k1lib, torch, math, gc, numpy as np; from functools import partial
import matplotlib.pyplot as plt
def allocated() -> int: return torch.cuda.memory_allocated()
class MemoryData: # handles hooks of 1 nn.Module
    def __init__(self, mProfiler, mS:k1lib.selector.ModuleSelector):
        self.mProfiler = mProfiler; self.mS = mS
        self.handles = k1lib.Object.fromDict({"fp":0,"f":0,"b":0})
        self.values = k1lib.Object.fromDict({"fp":0,"f":0,"b":0})
        self.hook()
    def hook(self):
        mS = self.mS; mP = self.mProfiler
        def hk(v, m, i, o=None): # v: type of hook
            gc.collect(); value = allocated() - mP.startMemory; self.values[v] += value
            if v == "f" or v == "b":
                mP.stepData.append([value, 0, mS.idx])
                # for the dashed line separating forward and backward
                if v == "b" and mP.startBackwardPoint is None: mP.startBackwardPoint = len(mP.stepData)
        self.handles.fp = mS.nn.register_forward_pre_hook  (partial(hk, "fp"))
        self.handles.f  = mS.nn.register_forward_hook      (partial(hk, "f"))
        self.handles.b  = mS.nn.register_full_backward_hook(partial(hk, "b"))
    def unhook(self):
        self.handles.fp.remove(); self.handles.f.remove(); self.handles.b.remove()
    def __getstate__(self):
        answer = dict(self.__dict__)
        del answer["mS"]; del answer["mProfiler"]; return answer
    def __setstate__(self, state): self.__dict__.update(dict(state))
    def __str__(self):
        fp = f"fp({fmt.size(self.values.fp)})".ljust(14)
        f  =  f"f({fmt.size(self.values.f)})" .ljust(13)
        b  =  f"b({fmt.size(self.values.b)})" .ljust(13)
        delta = f"delta({fmt.size(self.values.f - self.values.fp)})".ljust(17)
        return f"{b} {delta} {fp} {f}"
class MemoryProfiler(Callback):
    """Expected to be run only once only. If a new report for a new network
architecture is required, then create a new one. Example::

    l = k1lib.Learner.sample()
    l.cbs.add(Cbs.Profiler())
    # views graph and table
    l.Profiler.memory
    # views graph and table highlighted
    l.Profiler.memory.css("Linear")"""
    def startRun(self):
        if not hasattr(self, "selector"):
            self.selector = self.l.model.select("")
        for mS in self.selector.modules(): mS.data = MemoryData(self, mS)
        self.selector.displayF = lambda mS: (fmt.txt.red if "_memProf_" in mS else fmt.txt.identity)(mS.data)
        self.startMemory = allocated()
        self.stepData:List[Tuple[int, bool, int]] = [] # (bytes, css selected, mS.idx)
        self.startBackwardPoint = None
    def startStep(self): return True
    def endRun(self): self._updateLinState()
    def _run(self):
        """Runs everything"""
        with self.cbs.context(), self.cbs.suspendEval(), self.l.model.deviceContext():
            self.cbs.add(Cbs.Cuda()); self.l.run(1, 1)
        for m in self.selector.modules(): m.data.unhook()
    def _updateLinState(self):
        """Change linState, which is the graph's highlight"""
        @self.selector.apply
        def applyF(mS):
            for step in self.stepData:
                if step[2] == mS.idx: step[1] = "_memProf_" in mS
    def css(self, css:str):
        """Selects a small part of the network to highlight. See also: :mod:`k1lib.selector`."""
        self.selector.parse(k1lib.selector.preprocess(css, "_memProf_"))
        self._updateLinState(); print(self.__repr__())
        self.selector.clearProps(); self._updateLinState()
@k1lib.patch(MemoryProfiler)
def __repr__(self):
    plt.figure(dpi=120); plt.grid(True); plt.xlabel("Time")
    l, s, _ = self.stepData | cli.transpose() | cli.deref()
    label, l = fmt.sizeOf(l); plt.ylabel(label)
    k1lib.viz.plotSegments(range(len(l)), l, s)
    plt.axvline(self.startBackwardPoint, linestyle="--")
    ax = plt.gca(); ax.text(0.05, 0.05, "forward", transform=ax.transAxes)
    ax.text(0.95, 0.05, "backward", ha="right", transform=ax.transAxes); plt.show()
    c = self.selector.__repr__(intro=False).split("\n") | cli.tab() | cli.join("\n")
    return f"""MemoryProfiler (params: {fmt.item(self.l.model.nParams)}):\n{c}

Can...
- mp.css("..."): highlights a particular part of the network
- mp.selector: to get internal k1lib.selector.ModuleSelector object"""


# AUTOGENERATED FILE! PLEASE DON'T EDIT
from .callbacks import Callback, Callbacks, Cbs
import k1lib, numpy as np, time
import matplotlib.pyplot as plt
from typing import Callable
import k1lib.cli as cli
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["Landscape"]
spacing = 0.35 # orders of magnitude
offset = -2 # orders of magnitude shift
res = 20 # resolution
scales = 10**(np.array(range(8))*spacing + offset)
scales = [round(scale, 3) for scale in scales]
scales
F = Callable[["k1lib.Learner"], float]
@k1lib.patch(Cbs)
class Landscape(Callback):
    " "
    def __init__(self, propertyF:F, name:str=None):
        """Plots the landscape of the network.

:param propertyF: a function that takes in :class:`k1lib.Learner` and outputs the
    desired float property

.. warning::

    Remember to detach anything you get from :class:`k1lib.Learner` in your
    function, or else you're gonna cause a huge memory leak.
"""
        super().__init__(); self.propertyF = propertyF; self.suspended = True
        self.name = name or self.name; self.order = 23; self.parent:Callback = None
    def startRun(self): self.originalParams = self.l.model.exportParams()
    def endRun(self): self.l.model.importParams(self.originalParams)
    def startPass(self):
        next(self.iter)
        for param, og, v1, v2 in zip(self.l.model.parameters(), self.originalParams, *self.vs):
            param.data = og + self.x * v1 + self.y * v2
    def endLoss(self):
        prop = self.propertyF(self.l)
        self.zs[self.ix, self.iy] = prop if prop == prop else 0 # check for nan
        if self.l.batch % 10: print(f"\rProgress: {round(100*(self.ix+self.iy/res)/res)}%, {round(time.time()-self.beginTime)}s      ", end="")
    def startBackward(self): return True
    def startStep(self): return True
    def startZeroGrad(self): return True
    def __iter__(self):
        """This one is the "core running loop", if you'd like to say so. Because
this needs to be sort of event-triggered (by checkpoint "startPass"), so kinda have
to put this into an iterator so that it's not the driving thread."""
        self.zss = [] # debug data
        for i, (scale, ax) in enumerate(zip(scales, self.axes)):
            a = torch.linspace(-scale, scale, res)
            xs, ys = np.meshgrid(a, a); self.zs = np.empty((res, res))
            xs = torch.tensor(xs); ys = torch.tensor(ys)
            for ix in range(res):
                for iy in range(res):
                    self.x = xs[ix, iy]; self.y = ys[ix, iy]
                    self.ix, self.iy = ix, iy; yield True
            self.zs[self.zs == float("inf")] = 0
            ax.plot_surface(xs, ys, self.zs, cmap=plt.cm.coolwarm)
            self.zss.append(self.zs)
            print(f"     {i+1}/8 Finished [{-scale}, {scale}] range              ", end="")
        raise k1lib.CancelRunException("Landscape finished")
    def plot(self):
        """Creates the landscapes and show plots"""
        self.suspended = False; self.iter = iter(self); self.beginTime = time.time()
        def inner():
            self.vs = [self.l.model.getParamsVector(), self.l.model.getParamsVector()]
            fig, axes = plt.subplots(2, 4, subplot_kw={"projection": "3d"}, figsize=(16, 8), dpi=120)
            self.axes = axes.flatten(); self.l.run(1000000)
        try:
            with self.cbs.suspendEval(), torch.no_grad(): inner()
        except: pass
        self.suspended = True; self.iter = None
    def __repr__(self):
        return f"""{super()._reprHead}, use...
- l.plot(): to plot everything
{super()._reprCan}"""


# AUTOGENERATED FILE! PLEASE DON'T EDIT
import k1lib, time, torch; from k1lib import fmt, cli
from k1lib.callbacks import Callback, Cbs
_ltime = 15; _lt1 = 8; _lt2 = 18
class TimeData:
    def __init__(self, tProfiler, mS:k1lib.selector.ModuleSelector):
        self.tProfiler = tProfiler; self.mS = mS
        self.startTime = None; self.time = 0; self.hook()
    def hook(self):
        def fpHk(m, i):
            if self.tProfiler.is_cuda: torch.cuda.synchronize()
            self.startTime = time.time()
        def fHk(m, i, o):
            if self.tProfiler.is_cuda: torch.cuda.synchronize()
            self.time += time.time() - self.startTime
        self.fpH = self.mS.nn.register_forward_pre_hook(fpHk)
        self.fH = self.mS.nn.register_forward_hook(fHk)
    def unhook(self):
        self.tProfiler.totalTime = max(self.tProfiler.totalTime, self.time)
        self.fpH.remove(); self.fH.remove()
    def __getstate__(self):
        answer = dict(self.__dict__)
        del answer["mS"]; del answer["tProfiler"]; return answer
    def __setstate__(self, state): self.__dict__.update(dict(state))
    def __str__(self):
        if self.time <= 1e-20: return ""
        a = f"{fmt.time(self.time)}".ljust(_ltime)
        b = f"{round(100 * self.time / self.tProfiler.totalTime)}%".rjust(_lt1)
        c = f"{round(100 * self.time / self.tProfiler.selectedMaxTime)}%".rjust(_lt2) if self.tProfiler.selectedMaxTime != None and "_timeProf_" in self.mS else ""
        return f"{a}{b}{c}"
class TimeProfiler(Callback):
    """Profiles execution time. Only measures forward times, as
backward times can't really be measured. Example::

    l = k1lib.Learner.sample()
    l.cbs.add(Cbs.Profiler())
    # views table
    l.Profiler.time
    # views table highlighted
    l.Profiler.time.css("#lin1")"""
    def startRun(self):
        if not hasattr(self, "selector"): # if no selectors found
            self.selector = self.l.model.select("")
        for m in self.selector.modules(): m.data = TimeData(self, m)
        self.selector.displayF = lambda m: (fmt.txt.red if "_timeProf_" in m else fmt.txt.identity)(m.data)
        self.totalTime = 0; self.selectedMaxTime = None
    def startStep(self): return True
    def _run(self):
        """Runs everything"""
        with self.cbs.context(), self.cbs.suspendEval():
            self.is_cuda = next(self.l.model.parameters()).is_cuda
            if self.is_cuda: self.cbs.add(Cbs.Cuda())
            else: self.cbs.add(Cbs.Cpu())
            self.l.run(1, 1)
        for m in self.selector.modules(): m.data.unhook()
    def css(self, css:str):
        """Selects a small part of the network to highlight. See also: :mod:`k1lib.selector`."""
        self.selector.parse(k1lib.selector.preprocess(css, "_timeProf_"))
        self.selectedMaxTime = 0
        for m in self.selector.modules():
            if "_timeProf_" in m:
                self.selectedMaxTime = max(self.selectedMaxTime, m.data.time)
        print(self.__repr__())
        self.selector.clearProps(); self.selectedMaxTime = None
    def __repr__(self):
        header = "time".ljust(_ltime) + "% total".rjust(_lt1) + ("% selected max" if self.selectedMaxTime != None else "").rjust(_lt2)
        footer = ""
        if self.selectedMaxTime != None:
            b = f"{round(100 * self.selectedMaxTime / self.totalTime)}%".rjust(_lt1, " ")
            st = f"{fmt.time(self.selectedMaxTime)}".rjust(_lt2)
            footer = ("Selected max", " " * _ltime + b + st)
        c = self.selector.__repr__(intro=False, header=header, footer=footer).split("\n") | cli.tab() | cli.join("\n")
        return f"""TimeProfiler ({"GPU" if self.is_cuda else "CPU"}):\n{c}

Caveats: This one's a bit stranger than memory and computation profilers
1. There is no "total" time (adding all times in all modules). There
    is total network execution time tho, which is just the time taken
    for the top level module to execute
2. "% selected max" column displays percentage of selected max, not
    percentage of total selected time, which may matter in your analysis

Can...
- tp.css("..."): highlights a particular part of the network
- tp.selector: to get internal k1lib.ModuleSelector object"""


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""All tools related to GenBank file format. Expected to use behind the
"gb" module name, like this::

    from k1lib.imports import *
    cat("abc.gb") | gb.feats()
"""
from k1lib import cli
from typing import Any, Union, List
__all__ = ["feats", "origin"]
class feats(cli.BaseCli):
    """Fetches features, each on a separate stream.
Example::

    cat("a.gb") | gb.feats()

Output example::

    [['     source          1..248956422',
      '                     /organism="Homo sapiens"',
      '                     /mol_type="genomic DNA"',
      '                     /db_xref="taxon:9606"',
      '                     /chromosome="1"'],
     ['     gene            11874..14409',
      '                     /gene="DDX11L1"',
      '                     /note="DEAD/H-box helicase 11 like 1 (pseudogene); Derived',
      '                     by automated computational analysis using gene prediction',
      '                     method: BestRefSeq."',
      '                     /pseudo',
      '                     /db_xref="GeneID:100287102"',
      '                     /db_xref="HGNC:HGNC:37102"']]
"""
    def __ror__(self, it):
        it = it | cli.grep("FEATURES", 0, 1e9).till("ORIGIN") | cli.rows()[1:-1]
        cache = []
        for line in it:
            if line[4:9] != "     ": # new section detected
                if len(cache) > 0: yield cache
                cache = []
            cache.append(line)
        if len(cache) > 0: yield iter(cache)
    @staticmethod
    def filt(*terms:str) -> cli.BaseCli:
        """Filters for specific terms in all the features texts. If there
are multiple terms, then filters for first term, then second, then third,
so the term's order might matter to you. Example::

    [['     source          1..248956422',
      '                     /organism="Homo sapiens"',
      '                     /mol_type="genomic DNA"',
      '                     /db_xref="taxon:9606"',
      '                     /chromosome="1"'],
     ['     gene            11874..14409',
      '                     /gene="DDX11L1"',
      '                     /note="DEAD/H-box helicase 11 like 1 (pseudogene); Derived',
      '                     by automated computational analysis using gene prediction',
      '                     method: BestRefSeq."',
      '                     /pseudo',
      '                     /db_xref="GeneID:100287102"',
      '                     /db_xref="HGNC:HGNC:37102"']] | gb.feats.filt("mol_type")

Output::

    [['     source          1..248956422',
      '                     /organism="Homo sapiens"',
      '                     /mol_type="genomic DNA"',
      '                     /db_xref="taxon:9606"',
      '                     /chromosome="1"']]
"""
        if len(terms) == 0: return cli.iden()
        if len(terms) > 1: return cli.deref() | cli.init.serial(*(feats.filt(term) for term in terms))
        return cli.toList().all() | cli.filt(lambda F: F | cli.grep(terms[0]) | cli.shape(0) > 0)
    @staticmethod
    def root() -> cli.BaseCli:
        """Gets root (top most unnamed tag) of a feature.
Example::

     ['     misc_RNA        complement(join(14362..14829,14970..15038,15796..15947,',
      '                     16607..16765,16858..17055,17233..17368,17606..17742,',
      '                     17915..18061,18268..18366,24738..24891,29321..29370))',
      '                     /gene="WASH7P"',
      '                     /gene_synonym="FAM39F; WASH5P"',
      '                     /product="WASP family homolog 7, pseudogene"',
      '                     /note="Derived by automated computational analysis using',
      '                     gene prediction method: BestRefSeq."',
      '                     /pseudo',
      '                     /transcript_id="NR_024540.1"',
      '                     /db_xref="GeneID:653635"',
      '                     /db_xref="HGNC:HGNC:38034"'] | feats.root()

Output::

    ['misc_RNA',
     ['complement(join(14362..14829,14970..15038,15796..15947,',
      '16607..16765,16858..17055,17233..17368,17606..17742,',
      '17915..18061,18268..18366,24738..24891,29321..29370))']]
"""
        return cli.apply(lambda x: x.strip()) | cli.breakIf(lambda x: x.startswith("/")) | ~cli.aS(lambda a, *b: [a.split(" ")[0], [" ".join(a.split(" ")[1:]).strip(), *b]])
    @staticmethod
    def tags(*tags:List[str]) -> cli.BaseCli:
        """Grabs a list of tags.
Example::

    s = ['     misc_RNA        complement(join(14362..14829,14970..15038,15796..15947,',
         '                     16607..16765,16858..17055,17233..17368,17606..17742,',
         '                     17915..18061,18268..18366,24738..24891,29321..29370))',
         '                     /gene="WASH7P"',
         '                     /gene_synonym="FAM39F; WASH5P"',
         '                     /product="WASP family homolog 7, pseudogene"',
         '                     /note="Derived by automated computational analysis using',
         '                     gene prediction method: BestRefSeq."',
         '                     /pseudo',
         '                     /transcript_id="NR_024540.1"',
         '                     /db_xref="GeneID:653635"',
         '                     /db_xref="HGNC:HGNC:38034"']
    s | feats.tags()

Output::

    [['gene', 'WASH7P'],
     ['gene_synonym', 'FAM39F; WASH5P'],
     ['product', 'WASP family homolog 7, pseudogene'],
     ['note',
      'Derived by automated computational analysis using gene prediction method: BestRefSeq.'],
     ['pseudo', ''],
     ['transcript_id', 'NR_024540.1'],
     ['db_xref', 'GeneID:653635'],
     ['db_xref', 'HGNC:HGNC:38034']]

With filters::

    # returns [['gene', 'WASH7P'], ['db_xref', 'HGNC:HGNC:38034'], ['organism', '']]
    s | feats.tags("gene", "db_xref", "organism")
"""
        f = cli.op().strip().all() | cli.grep("^/", sep=True).till() | cli.join(" ").all()\
            | (cli.op().split("=") | ~cli.aS(lambda a, *b: [a[1:], b[0][1:-1] if len(b) > 0 else ""])).all()
        if len(tags) == 0: return f | cli.deref()
        def g(it):
            d = it | f | cli.toDict()
            return [[tag, d[tag] if tag in d else ""] for tag in tags]
        return cli.aS(g)
class origin(cli.BaseCli):
    """Return the origin section of the genbank file.
Example::

    # returns single fasta string
    cat("a.gb") | gb.origin()
"""
    def __ror__(self, it):
        return it | cli.grep("ORIGIN", 0, 1e9) | ~cli.head(1) | cli.op().strip().all()\
        | cli.op().split(" ").all() | cli.cut()[1:] | cli.join("").all()\
        | cli.op().replace("/", "").all() | cli.join("")


# AUTOGENERATED FILE! PLEASE DON'T EDIT
import numpy as np, matplotlib.pyplot as plt, matplotlib as mpl
from typing import Any, List, Union, Tuple
__all__ = ["perlin3d"]
def interpolant(t):
    return t*t*t*(t*(t*6 - 15) + 10)
def perlin3d(shape=(100, 100, 100), res=(2, 2, 2), tileable=(False, False, False), interpolant=interpolant):
    """Generate a 3D numpy array of perlin noise. Not my code! All credits go
to the author of this library: https://github.com/pvigier/perlin-numpy

:param shape: The shape of the generated array (tuple of three ints).
    This must be a multiple of res.
:param res: The number of periods of noise to generate along each
    axis (tuple of three ints). Note shape must be a multiple
    of res.
:param tileable: If the noise should be tileable along each axis
    (tuple of three bools). Defaults to (False, False, False).
:param interpolant: The interpolation function, defaults to
    t*t*t*(t*(t*6 - 15) + 10).
:return: A numpy array of shape shape with the generated noise.
:raises ValueError: If shape is not a multiple of res."""
    delta = (res[0] / shape[0], res[1] / shape[1], res[2] / shape[2])
    d = (shape[0] // res[0], shape[1] // res[1], shape[2] // res[2])
    grid = np.mgrid[0:res[0]:delta[0],0:res[1]:delta[1],0:res[2]:delta[2]]
    grid = np.mgrid[0:res[0]:delta[0],0:res[1]:delta[1],0:res[2]:delta[2]]
    grid = grid.transpose(1, 2, 3, 0) % 1
    # Gradients
    theta = 2*np.pi*np.random.rand(res[0] + 1, res[1] + 1, res[2] + 1)
    phi = 2*np.pi*np.random.rand(res[0] + 1, res[1] + 1, res[2] + 1)
    gradients = np.stack(
        (np.sin(phi)*np.cos(theta), np.sin(phi)*np.sin(theta), np.cos(phi)),
        axis=3
    )
    if tileable[0]:
        gradients[-1,:,:] = gradients[0,:,:]
    if tileable[1]:
        gradients[:,-1,:] = gradients[:,0,:]
    if tileable[2]:
        gradients[:,:,-1] = gradients[:,:,0]
    gradients = gradients.repeat(d[0], 0).repeat(d[1], 1).repeat(d[2], 2)
    g000 = gradients[    :-d[0],    :-d[1],    :-d[2]]
    g100 = gradients[d[0]:     ,    :-d[1],    :-d[2]]
    g010 = gradients[    :-d[0],d[1]:     ,    :-d[2]]
    g110 = gradients[d[0]:     ,d[1]:     ,    :-d[2]]
    g001 = gradients[    :-d[0],    :-d[1],d[2]:     ]
    g101 = gradients[d[0]:     ,    :-d[1],d[2]:     ]
    g011 = gradients[    :-d[0],d[1]:     ,d[2]:     ]
    g111 = gradients[d[0]:     ,d[1]:     ,d[2]:     ]
    # Ramps
    n000 = np.sum(np.stack((grid[:,:,:,0]  , grid[:,:,:,1]  , grid[:,:,:,2]  ), axis=3) * g000, 3)
    n100 = np.sum(np.stack((grid[:,:,:,0]-1, grid[:,:,:,1]  , grid[:,:,:,2]  ), axis=3) * g100, 3)
    n010 = np.sum(np.stack((grid[:,:,:,0]  , grid[:,:,:,1]-1, grid[:,:,:,2]  ), axis=3) * g010, 3)
    n110 = np.sum(np.stack((grid[:,:,:,0]-1, grid[:,:,:,1]-1, grid[:,:,:,2]  ), axis=3) * g110, 3)
    n001 = np.sum(np.stack((grid[:,:,:,0]  , grid[:,:,:,1]  , grid[:,:,:,2]-1), axis=3) * g001, 3)
    n101 = np.sum(np.stack((grid[:,:,:,0]-1, grid[:,:,:,1]  , grid[:,:,:,2]-1), axis=3) * g101, 3)
    n011 = np.sum(np.stack((grid[:,:,:,0]  , grid[:,:,:,1]-1, grid[:,:,:,2]-1), axis=3) * g011, 3)
    n111 = np.sum(np.stack((grid[:,:,:,0]-1, grid[:,:,:,1]-1, grid[:,:,:,2]-1), axis=3) * g111, 3)
    # Interpolation
    t = interpolant(grid)
    n00 = n000*(1-t[:,:,:,0]) + t[:,:,:,0]*n100
    n10 = n010*(1-t[:,:,:,0]) + t[:,:,:,0]*n110
    n01 = n001*(1-t[:,:,:,0]) + t[:,:,:,0]*n101
    n11 = n011*(1-t[:,:,:,0]) + t[:,:,:,0]*n111
    n0 = (1-t[:,:,:,1])*n00 + t[:,:,:,1]*n10
    n1 = (1-t[:,:,:,1])*n01 + t[:,:,:,1]*n11
    return ((1-t[:,:,:,2])*n0 + t[:,:,:,2]*n1)


# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""
This is for functions that are .sam or .bam related
"""
from k1lib import cli; import k1lib
__all__ = ["cat", "header", "flag"]
settings = k1lib.Settings()
k1lib.settings.cli.add("sam", settings, "from k1lib.cli.sam module");
catF = lambda header: cli.applyS(lambda bamFile: None | cli.cmd(f"samtools view {'-h' if header else ''} {bamFile}") | cli.table("\t"))
def cat(bamFile:str=None, header:bool=True):
    """Get sam file outputs from bam file.
Example::

    sam.cat("file.bam") | display()
    "file.bam" | sam.cat(header=False) | display()

:param header: whether to include headers or not"""
    return catF(header)(bamFile) if bamFile is not None else catF(header)
settings.add("header", k1lib.Settings()
             .add("short", ["qname", "flag", "rname", "pos", "mapq", "cigar", "rnext", "pnext", "tlen", "seq", "qual"])
             .add("long", ["Query template name", "Flags", "Reference sequence name", "Position", "Mapping quality", "CIGAR string", "Rname of next read", "Position of next read", "Template length", "Sequence", "Quality"]), "sam headers")
class header(cli.BaseCli):
    def __init__(self, long=True):
        """Adds a header to the table.
Example::

    sam.cat("file.bam") | sam.header() | display()

You can change the header labels like this::

    settings.cli.sam.header.long = ["Query template name", ...]

:param long: whether to use a long descriptive header, or a short one"""
        super().__init__(); self.long = long
    def __ror__(self, it):
        return it | cli.insert(settings.header.long if self.long else settings.header.short)
settings.add("flags", ['PAIRED', 'PROPER_PAIR', 'UNMAP', 'MUNMAP', 'REVERSE', 'MREVERSE', 'READ1', 'READ2', 'SECONDARY', 'QCFAIL', 'DUP', 'SUPPLEMENTARY'], "list of flags")
class flag(cli.bindec):
    def __init__(self, f=None):
        """Decodes flags attribute.
Example::

    # returns ['PAIRED', 'UNMAP']
    5 | flag()
    # returns 'PAIRED, UNMAP'
    5 | flag(cli.join(", "))

You'll mostly use this in this format::

    sam.cat("file.bam", False) | apply(sam.flag(), 1) | display()

You can change the flag labels like this::

    settings.cli.sam.flags = ["paired", ...]

:param f: transform function fed into :class:`~k1lib.cli.utils.bindec`, defaulted
    to `join(", ")`"""
        super().__init__(k1lib.settings.cli.sam.flags, f or cli.join(", "))